
def val_dets_3_7(val_el):
    """
    val_el is the element under value e.g. Constant (3.8+) or Str, Num (<3.8)
    """
    raw_val = val_el.get('s')
    if raw_val:
        val = raw_val
        needs_quoting = True
    else:
        raw_val = val_el.get('n')
        if raw_val:
            val = float(raw_val)  ## 1 and 1.0 are the same val when a key in a dict
            needs_quoting = False
        else:
            return None
    return val, needs_quoting

def val_dets_3_8(val_el):
    """
    As per 3_7 version
    """
    const_type = val_el.get('type')
    raw_val = val_el.get('value')
    if const_type == 'int':
        val = int(raw_val)
        needs_quoting = False
    elif const_type == 'float':
        val = float(raw_val)
        needs_quoting = False
    elif const_type == 'str':
        val = raw_val
        needs_quoting = True
    else:
        return None
    return val, needs_quoting

## nums ******************************

def assigned_num_els_from_block_3_7(block_el):
    num_els = block_el.xpath('descendant-or-self::Assign/value/Num')
    return num_els

def assigned_num_els_from_block_3_8(block_el):
    val_els = block_el.xpath('descendant-or-self::Assign/value')
    num_els = []
    for val_el in val_els:
        constant_els = val_el.xpath('Constant')
        if len(constant_els) != 1:
            continue
        constant_el = constant_els[0]
        if constant_el.get('type') in ('int', 'float'):
            num_els.append(constant_el)
    return num_els

def num_str_from_val_3_7(val_el):
    positive_num_els = val_el.xpath('Num')
    if len(positive_num_els) == 1:
        num = positive_num_els[0].get('n')
    else:
        sub_els = val_el.xpath('UnaryOp/op/USub')
        if not sub_els:
            return None
        negative_num_els = val_el.xpath('UnaryOp/operand/Num')
        if len(negative_num_els) != 1:
            return None
        pos_num = negative_num_els[0].get('n')
        num = f'-{pos_num}'
    return num

def num_str_from_val_3_8(val_el):
    positive_num_els = val_el.xpath('Constant')
    if len(positive_num_els) == 1:
        positive_num_el = positive_num_els[0]
        val = positive_num_el.get('value')
        if positive_num_el.get('type') not in ('int', 'float'):
            return None
        num = val
    else:
        sub_els = val_el.xpath('UnaryOp/op/USub')
        if not sub_els:
            return None
        negative_num_els = val_el.xpath('UnaryOp/operand/Constant')
        if len(negative_num_els) != 1:
            return None
        negative_num_el = negative_num_els[0]
        val = negative_num_el.get('value')
        if negative_num_el.get('type') not in ('int', 'float'):
            return None
        num = f'-{val}'
    return num

def num_str_from_el_3_7(comparison_el):
    num = comparison_el.get('n')
    if not num:
        return None
    return num

def num_str_from_el_3_8(comparison_el):
    val = comparison_el.get('value')
    if comparison_el.get('type') in ('int', 'float'):
        num = val
    else:
        num = None
    return num

## strs ******************************

def str_from_el_3_7(el):
    string = el.get('s')
    if not string:
        return None
    return string

def str_from_el_3_8(el):
    val = el.get('value')
    if el.get('type') == 'str':
        string = val
    else:
        string = None
    return string

def assigned_str_els_from_block_3_7(block_el):
    str_els = block_el.xpath('descendant-or-self::Assign/value/Str')
    return str_els

def assigned_str_els_from_block_3_8(block_el):
    assign_val_els = block_el.xpath('descendant-or-self::Assign/value')
    str_els = []
    for assign_val_el in assign_val_els:
        assign_str_els = assign_val_el.xpath('Constant')
        if len(assign_str_els) != 1:
            continue
        assign_str_el = assign_str_els[0]
        if assign_str_el.get('type') == 'str':
            str_els.append(assign_str_el)
    return str_els

def str_els_from_block_3_7(block_el):
    str_els = block_el.xpath('descendant-or-self::Str')
    return str_els

def str_els_from_block_3_8(block_el):
    constant_els = block_el.xpath('descendant-or-self::Constant')
    str_els = [constant_el for constant_el in constant_els
        if constant_el.get('type') == 'str']
    return str_els

## dict keys ******************************

def dict_key_from_subscript_3_7(subscript_el):
    key_els = subscript_el.xpath(
        'slice/Index/value/Num | slice/Index/value/Str')
    if len(key_els) != 1:
        return None
    key_el = key_els[0]
    return val_dets_3_7(val_el=key_el)

def dict_key_from_subscript_3_8(subscript_el):
    key_els = subscript_el.xpath('slice/Index/value/Constant')
    if len(key_els) != 1:
        return None
    key_el = key_els[0]
    return val_dets_3_8(val_el=key_el)

## other ******************************

def _get_var_plus_equalled_all(el):
    plus_equalled_els = el.xpath('descendant-or-self::AugAssign')
    if len(plus_equalled_els) != 1:
        return None
    plus_equalled_el = plus_equalled_els[0]
    target_els = plus_equalled_el.xpath('target/Name')
    if len(target_els) != 1:
        return None
    target = target_els[0].get('id')  ## e.g. 'i' or 'counter' etc
    add_sub_els = plus_equalled_el.xpath('op/Add | op/Sub')  ## only interested in Add / Sub
    if len(add_sub_els) != 1:
        return None
    return target, plus_equalled_el

def _get_var_plus_equalled_3_7(el):
    ok_num = '1'
    res = _get_var_plus_equalled_all(el)
    if res is None:
        return None
    target, plus_equalled_el = res
    num_els = plus_equalled_el.xpath('value/Num')
    if len(num_els) != 1:
        return None
    num = num_els[0].get('n')
    if num != ok_num:
        return None
    var_plus_equalled = target
    return var_plus_equalled

def _get_var_plus_equalled_3_8(el):
    ok_num = '1'
    res = _get_var_plus_equalled_all(el)
    if res is None:
        return None
    target, plus_equalled_el = res
    num_els = plus_equalled_el.xpath('value/Constant')
    if len(num_els) != 1:
        return None
    num_el = num_els[0]
    val = num_el.get('value')
    if num_el.get('type') not in ('int', 'float'):
        return None
    if val != ok_num:
        return None
    var_plus_equalled = target
    return var_plus_equalled

def _get_var_equal_plussed_all(el):
    assign_els = el.xpath('descendant-or-self::Assign')
    if len(assign_els) != 1:
        return None
    assign_el = assign_els[0]
    target_name_els = assign_el.xpath('targets/Name')
    if len(target_name_els) != 1:
        return None
    var_name = target_name_els[0].get('id')
    binop_els = assign_el.xpath('value/BinOp')
    if len(binop_els) != 1:
        return None
    bin_op_el = binop_els[0]
    left_val_els = bin_op_el.xpath('left/Name')
    if len(left_val_els) != 1:
        return None
    left_val = left_val_els[0].get('id')
    if left_val != var_name:  ## we want i = i ... + 1
        return None
    add_els = bin_op_el.xpath('op/Add')
    if len(add_els) != 1:
        return None
    return var_name, bin_op_el

def _get_var_equal_plussed_3_7(el):
    ok_val = '1'
    res = _get_var_equal_plussed_all(el)
    if res is None:
        return None
    var_name, bin_op_el = res
    right_els = bin_op_el.xpath('right/Num')
    if len(right_els) != 1:
        return None
    right_val = right_els[0].get('n')  ## ... + 1
    if right_val != ok_val:
        return None
    var_equal_plussed = var_name
    return var_equal_plussed

def _get_var_equal_plussed_3_8(el):
    ok_val = '1'
    res = _get_var_equal_plussed_all(el)
    if res is None:
        return None
    var_name, bin_op_el = res
    right_els = bin_op_el.xpath('right/Constant')
    if len(right_els) != 1:
        return None
    right_el = right_els[0]
    right_val = right_el.get('value')  ## ... + 1
    if right_el.get('type') not in ('int', 'float'):
        return None
    if right_val != ok_val:
        return None
    var_equal_plussed = var_name
    return var_equal_plussed

def get_danger_status_3_7(child_el):
    if (child_el.tag == 'NameConstant'
            and child_el.get('value') in ['True', 'False']):
        danger_status = 'Boolean'
    elif child_el.tag == 'Num' and child_el.get('n'):
        danger_status = 'Number'
    else:
        danger_status = None
    return danger_status

def get_danger_status_3_8(child_el):
    if child_el.tag == 'Constant':
        val = child_el.get('value')
        if val in ['True', 'False']:
            danger_status = 'Boolean'
        else:
            try:
                float(val)
            except TypeError:
                danger_status = None
            else:
                danger_status = 'Number'
    else:
        danger_status = None
    return danger_status

def get_docstring_from_value_3_7(first_value_el):
    if first_value_el.tag != 'Str':
        docstring = None
    else:
        docstring = first_value_el.get('s')
    return docstring

def get_docstring_from_value_3_8(first_value_el):
    if first_value_el.tag != 'Constant':
        docstring = None
    elif first_value_el.get('type') != 'str':
        docstring = None
    else:
        docstring = first_value_el.get('value')
    return docstring

def get_slice_dets_3_7(assign_subscript_el):
    """
    Get slice dets e.g. '[2]', '[-1]' or '["NZ"]'. Only interested in numbers or
    strings. If the person is doing something funky they'll miss out on learning
    about names and values.
    """
    val_els = assign_subscript_el.xpath('slice/Index/value')
    if len(val_els) != 1:
        return None
    val_el = val_els[0]
    num_str = num_str_from_val_3_7(val_el)
    if num_str:
        slice_dets = slice_dets = f'[{num_str}]'
    else:
        slice_str_els = assign_subscript_el.xpath('slice/Index/value/Str')
        if slice_str_els:
            slice_str = slice_str_els[0].get('s')
            slice_dets = f'["{slice_str}"]'
        else:
            raise Exception("Unable to get slice_dets")
    return slice_dets

def get_slice_dets_3_8(assign_subscript_el):
    """
    As for get_slice_dets_3_7.
    """
    val_els = assign_subscript_el.xpath('slice/Index/value')
    if len(val_els) != 1:
        return None
    val_el = val_els[0]
    num_str = num_str_from_val_3_8(val_el)
    if num_str:
        slice_dets = slice_dets = f'[{num_str}]'
    else:
        val_constant_els = val_el.xpath('Constant')
        if len(val_constant_els) != 1:
            raise Exception("Unable to get slice_dets")
        val_constant_el = val_constant_els[0]
        val_type = val_constant_el.get('type')
        if val_type == 'str':
            slice_str = val_constant_el.get('value')
            slice_dets = f'["{slice_str}"]'
        else:
            raise Exception("Unable to get slice_dets")
    return slice_dets

def _get_nt_lbl_flds_any(assign_block_el, tag='Constant', id_attr='value'):
    """
    Either two strs
      e.g. 'DataDets', 'a, b, c'
    OR a Constant/Str and list of n Constant/Strs
      e.g. 'DataDets', ['a', 'b', 'c']
    OR a Constant/Str and a tuple of n Constants/Strs
      e.g. 'DataDets', ('a', 'b', 'c')
    """
    args_els = assign_block_el.xpath('value/Call/args')
    if len(args_els) != 1:
        raise Exception("Expected only one args el for named tuple")
    args_el = args_els[0]
    label_el, fields_el = args_el.getchildren()
    label = label_el.get(id_attr)
    field_el_type = fields_el.tag
    if field_el_type == tag:
        fields_str = fields_el.get(id_attr)
        fields_list = [field.strip() for field in fields_str.split(',')]
    elif field_el_type in ('Tuple', 'List'):
        field_els = fields_el.xpath(f'elts/{tag}')
        fields_list = [el.get(id_attr) for el in field_els]
    return label, fields_list

def get_nt_lbl_flds_3_7(assign_block_el):
    return _get_nt_lbl_flds_any(assign_block_el, tag='Str', id_attr='s')

def get_nt_lbl_flds_3_8(assign_block_el):
    return _get_nt_lbl_flds_any(
        assign_block_el, tag='Constant', id_attr='value')

def get_slice_n_3_7(assign_el):
    slice_n = assign_el.xpath(
        'value/Subscript/slice/Index/value/Num')[0].get('n')
    return slice_n

def get_slice_n_3_8(assign_el):
    val_els = assign_el.xpath('value/Subscript/slice/Index/value/Constant')
    val_el = val_els[0]
    if val_el.get('type') in ('int', 'float'):
        slice_n = val_el.get('value')
    else:
        raise TypeError("slice index value not an int or a float - actual type "
            f"'{val_el.get('type')}'")
    return slice_n

def get_str_els_being_combined_3_7(block_el):
    str_els_being_combined = block_el.xpath(
        'descendant-or-self::BinOp/left/Str')
    return str_els_being_combined

def get_str_els_being_combined_3_8(block_el):
    left_str_els = block_el.xpath('descendant-or-self::BinOp/left/Constant')
    str_els_being_combined = []
    for left_str_el in left_str_els:
        if left_str_el.get('type') == 'str':
            str_els_being_combined.append(left_str_el)
    return str_els_being_combined



from .. import conf
from ..gen_utils import layout_comment as layout
from ..helpers import snippet_str_help

@snippet_str_help()
def lambda_advice(snippet):
    """
    Look for use of lambda and given general advice on when / how to use.
    """
    if not 'lambda' in snippet:
        return None

    title = layout("""\
    ### Using `lambda`
    """)
    brief_msg = layout("""\

    Lambdas are commonly overused so it is worth reading
    <https://treyhunner.com/2018/09/stop-writing-lambda-expressions/>. Having
    said this, lambdas used appropriately in sorting as key are idiomatic and
    readable as a result.
    """)
    main_msg = (
        layout("""\

        Lambda functions are anonymous functions. They have no name or
        documentation so should only be used where their brevity pros outweigh
        their readability cons.

        In their favour, lambdas are idiomatic when used as the key in sorting
        operations. They are also commonly relied upon when using libraries like
        Pandas (albeit not always wisely).

        Sometimes the alternatives are arguably worse. Consider the following
        alternatives - in this case it may be difficult to improve on the
        lambda:

        #### Use `lambda`

        Using a lambda creates one idiomatic line which obviously sorts by the
        country and city keys in that order
        """)
        +
        layout("""\
        addresses.sort(key=lambda address: (address['country'], address['city']))
        """, is_code=True)
        +
        layout("""\

        #### Alternative 1) use `operator` module

        Using `itemgetter` requires an extra import and it's not as idiomatic.
        Using `operator` functions may become more idiomatic but probably not.
        """)
        +
        layout("""\
        from operator import itemgetter

        addresses.sort(key=(itemgetter('country'), itemgetter('city')))
        """, is_code=True)
        +
        layout("""\

        #### Alternative 2) define a conventional function

        Defining a named function adds many more lines of code and may break up
        the flow of overall program logic. Often not worth it when a very simple
        function.
        """)
        +
        layout("""\
        def country_city(address):
            return (address['country'], address['city'])

        addresses.sort(key=country_city)
        """, is_code=True)
        +
        layout("""\

        Lambdas are a clear mistake when they add nothing or when they become
        too complex.

        Lambdas add nothing apart from confusion and noise when they merely
        apply a function. In the following examples both lines in each pair
        achieve the same result:
        """)
        +
        layout("""\
        words.sort(key=lambda word: len(word))
        words.sort(key=len)

        numbers.sort(key=lambda num: abs(num))
        numbers.sort(key=abs)
        """, is_code=True)
        +
        layout("""\

        In the following example a simple, clearly documented (and testable)
        function is a much better option than a lambda. Compare:
        """)
        +
        layout("""\
        water_iot_vals.sort(
            key=lambda x: min((sum(x) + min(x)) / max(x), 1) + min(x) ** 2)
        """, is_code=True)
        +
        layout("""\
        against the more readable, maintainable, and testable:
        """)
        +
        layout('''\
        def water_quality(vals):
            """
            Reading rank is based on ... etc etc

            The Tajik adjustment was favoured over the more traditional Core
            Value adjustment because ...

            See https://...calculating-clark-coefficients

            :param list vals: ...
            :return: water quality rating
            :rtype: float
            """
            raw_clark_coeff = min((sum(vals) + min(vals)) / max(vals)
            corrected_clark_coeff = min(raw_clark_coeff, 1)
            tajik_adjustment = min(vals) ** 2
            adjusted_clark_coeff = corrected_clark_coeff + tajik_adjustment
            return adjusted_clark_coeff

        water_iot_vals.sort(key=water_quality)
        ''', is_code=True)
        + layout("""\

        The main rule with lambdas is to avoid them unless they make the code
        more readable. And if you must use them, remember that you can still use
        names that convey some meaning - for example `s` for string, `t` for
        tuple, `d` for dict, `l` for list, `nt` for namedtuple, `row` for a row
        etc. They will make the logic more intelligible and make mistakes more
        obvious. For example, x[-1] might be what was intended but it would be
        easier to tell if we saw:

        `lambda l: l[-1]`  ## obviously the last item of a list

        or `lambda s: s[-1]`  ## obviously the last character of a string

        or `lambda d: d[-1]`  ## probably a mistake unless -1 is a key

        Of course, longer names should be used if they help.
        """)
    )
    extra = layout("""\

    Trey Hunner's "Overusing lambda expressions in Python"
    <https://treyhunner.com/2018/09/stop-writing-lambda-expressions/> is well
    worth reading. The position taken in SuperHELP differs in some regards but
    has been influenced by Trey's thinking.

    If using functions from the `operator` module (`itemgetter`, `attrgetter`,
    and `methodcaller`) becomes more common and idiomatic then it could be time
    to substantially reduce the usage of `lambda`.
    """)

    message = {
        conf.BRIEF: title + brief_msg,
        conf.MAIN: title + main_msg,
        conf.EXTRA: extra,
    }
    return message
    


from . import cli_colour, cli_conf, cli_utils
from ... import conf

def _get_vertical_padding_line(length):
    return length * ' '

def _get_true_width(text):
    """
    All the characters between the start STX and end ETX transmission characters
    count as one only per block.
    """
    n_chars = 0
    collect = True
    for char in text:
        if char == cli_conf.STX:
            collect = False
            n_chars += 1  ## just counting as 1
            continue
        elif char == cli_conf.ETX:
            collect = True
            continue
        elif collect:
            n_chars += 1
        elif not collect:
            continue
        else:
            raise Exception("Logically should never be here")
    true_width = n_chars
    return true_width

def _visibly_centred_text(text):
    """
    Centre based on characters that can be seen and that actually contribute to
    width.
    """
    true_width = _get_true_width(text)
    padding_needed = (cli_conf.TERMINAL_WIDTH - true_width)
    if padding_needed <= 0:
        true_centred_text = text
    else:
        even_padding = int(padding_needed / 2)
        initial_tot_width = (even_padding + true_width + even_padding)
        extra_needed = cli_conf.TERMINAL_WIDTH - initial_tot_width
        if extra_needed == 0:
            true_centred_text = (
                (even_padding * ' ') + text + (even_padding * ' '))
        elif extra_needed == 1:
            true_centred_text = (
                ((even_padding + 1) * ' ') + text + (even_padding * ' '))
        else:
            raise Exception(f"Unexpected extra_needed: {extra_needed}")
    return true_centred_text

def h(text, level):
    level_colour = cli_colour.LEVEL2COLOUR.get(level)
    bold = False
    if level <= 2:
        vertical_padding_line = _get_vertical_padding_line(
            length=cli_conf.TERMINAL_WIDTH)
        if len(text) < cli_conf.TERMINAL_WIDTH:
            text = text.center(cli_conf.TERMINAL_WIDTH)
        else:
            split_text = [subtext.replace('\n', '').strip()
                for subtext in text.split(conf.FORCE_SPLIT)]
            centred_subtext = [
                _visibly_centred_text(sub_text) for sub_text in split_text]
            text = '\n'.join(centred_subtext)
    else:
        if level == 3:
            lstar_padding = '**** '
            rstar_padding = ' ****'
        else:
            lstar_padding, rstar_padding = '', ''
        text = f"{lstar_padding}{text}{rstar_padding} "  ## extra space at end so right side of reversed background has a space between it and the text
        vertical_padding_line = _get_vertical_padding_line(length=len(text))
    if level <= 2:
        text = f"{vertical_padding_line}\n{text}\n{vertical_padding_line}"
        bold = True
    if level == 1:
        text = f"\n{text}"
    return cli_colour.colourise(
        text, level_colour, reverse=True, bold=bold) + '\n'

def h1(s, **_kwargs):
    return h(s, level=1)

def h2(s, **_kwargs):
    return h(s, level=2)

def h3(s, **_kwargs):
    return h(s, level=3)

def h4(s, **_kwargs):
    return h(s, level=4)

def h5(s, **_kwargs):
    return h(s, level=5)

def p(text, **_kwargs):
    return cli_colour.colourise(text.strip('\n') + '\n', cli_colour.TEXT)

def a(text, **_kwargs):
    return cli_colour.colourise_low_vis(text)

def hr(_text, **kw):
    """
    Horizontal line (rule) that has colour of appropriate heading level.
    """
    nesting_level = kw.get('nesting_level', 1)
    indent = (nesting_level - 1) * cli_conf.LEFT_INDENT
    colour = cli_colour.LEVEL2COLOUR[nesting_level]
    start = end = cli_colour.colourise(cli_conf.HR_ENDS, colour)
    return cli_colour.colourise_low_vis(
        f"\n{indent}{start}{cli_conf.HR_MARKER}{end}{indent}\n")

def code(text, from_fenced_block=None, **kw):
    """
    md code AND ``` style fenced raw code ends here
    """
    if not from_fenced_block:
        text = ('\n' + text).replace('\n    ', '\n')[1:]
    n_text_lines = len(text.split('\n'))
    # funny: ":-" confuses the tokenizer. replace/backreplace:
    raw_code = text.replace(':-', '\x01--')
    raw_code = raw_code.replace('## &gt;&gt;&gt;', '## >>>')
    text = cli_utils.style_ansi(raw_code)
    # unnested level has indent of N_LEFT_INDENT, use it for fenced
    indent = ' ' * kw.get('nesting_level', cli_conf.N_LEFT_INDENT)
    # if from_fenced_block: ... WE treat equal.
    # shift to the far left, no matter the indent (screen space matters):
    firstl = text.split('\n')[0]
    n_spaces2delete = len(firstl) - len(firstl.lstrip())
    spaces2delete = ' ' * n_spaces2delete
    width = len(str(n_text_lines))  ## e.g. 3 so we can cope with line number 100 onwards
    width_of_line_num_1 = 1
    missing_indents = width - width_of_line_num_1  ## so if 3 wide to handle 100+ we need 2
    indent_for_first_line_num = missing_indents * ' '
    text = (
            '\n'
            + indent_for_first_line_num
            + (f"\n{text}").replace(f"\n{spaces2delete}", '\n')[1:]
        )
    # we want an indent of one and low vis prefix. this does it:
    code_lines = text.splitlines()
    code_prefix = cli_colour.colourise_low_vis(cli_conf.CODE_PREFIX)
    empty = cli_colour.colourise('', cli_colour.CODE_COLOUR, no_reset=True)
    prefix = f"\n{indent}{code_prefix} {empty}"
    if code_lines[-1] == '\x1b[0m':
        code_lines.pop()
    code_str = prefix.join(code_lines)
    code_str = code_str.replace('\x01--', ':-')
    return code_str + '\n' + cli_colour.DEFAULT_ANSI_COLOUR_BYTE_STR



from collections import namedtuple

from . import ast_funcs, conf

AssignedNameDets = namedtuple('AssignedNameDets',
    'name_type, name_details, name_str, unpacking_idx')

def _std_name_dets_from_std_name_el(std_name_el, unpacking_idx=None):
    name_str = std_name_el.get('id')
    name_details = [name_str, ]
    name_type = conf.STD_NAME
    return AssignedNameDets(name_type, name_details, name_str, unpacking_idx)

def _std_name_dets_from_assign_el(assign_el):
    std_name_els = assign_el.xpath('targets/Name')
    if len(std_name_els) != 1:
        return None
    std_name_el = std_name_els[0]
    std_name_dets = _std_name_dets_from_std_name_el(std_name_el)
    return std_name_dets

def _obj_attr_name_dets_from_attribute_el(attribute_el, unpacking_idx=None):
    attribute_name = attribute_el.get('attr')
    obj_name_els = attribute_el.xpath('value/Name')
    if len(obj_name_els) != 1:
        return None
    obj_name_el = obj_name_els[0]
    obj_only_name = obj_name_el.get('id')
    name_str = f"{obj_only_name}.{attribute_name}"
    name_details = [obj_only_name, attribute_name]
    name_type = conf.OBJ_ATTR_NAME
    return AssignedNameDets(name_type, name_details, name_str, unpacking_idx)

def _obj_attr_name_dets_from_assign_el(assign_el):
    """
    The target name is the attribute of an object.
    """
    attribute_els = assign_el.xpath('targets/Attribute')
    if len(attribute_els) != 1:
        return None
    attribute_el = attribute_els[0]
    return _obj_attr_name_dets_from_attribute_el(attribute_el)

def _dict_key_name_dets_from_subscript_el(subscript_el, unpacking_idx=None):
    dict_name_els = subscript_el.xpath('value/Name')
    if len(dict_name_els) != 1:
        return None
    dict_name_el = dict_name_els[0]
    dict_name = dict_name_el.get('id')
    dict_key, needs_quoting = ast_funcs.dict_key_from_subscript(subscript_el)
    quoter = '"' if needs_quoting else ''
    name_str = f"{dict_name}[{quoter}{dict_key}{quoter}]"
    name_details = [dict_name, dict_key]
    name_type = conf.DICT_KEY_NAME
    return AssignedNameDets(name_type, name_details, name_str, unpacking_idx)

def _dict_key_name_dets_from_assign_el(assign_el):
    """
    The target name is the key of a dictionary.
    """
    subscript_els = assign_el.xpath('targets/Subscript')
    if len(subscript_els) != 1:
        return None
    subscript_el = subscript_els[0]
    return _dict_key_name_dets_from_subscript_el(subscript_el)

def _tuple_names_dets_from_assign_el(assign_el):
    """
    The target names are in a tuple.
    """
    tuple_els = assign_el.xpath('targets/Tuple')
    if len(tuple_els) != 1:
        return None
    tuple_el = tuple_els[0]
    tuple_elts = tuple_el.getchildren()[0]
    tuple_item_els = tuple_elts.getchildren()
    tuple_names_dets = []
    for unpacking_idx, tuple_item_el in enumerate(tuple_item_els):
        if tuple_item_el.tag == 'Name':
            std_name_dets = _std_name_dets_from_std_name_el(
                std_name_el=tuple_item_el, unpacking_idx=unpacking_idx)
            tuple_names_dets.append(std_name_dets)
        elif tuple_item_el.tag == 'Attribute':
            obj_attr_name_dets = _obj_attr_name_dets_from_attribute_el(
                attribute_el=tuple_item_el, unpacking_idx=unpacking_idx)
            tuple_names_dets.append(obj_attr_name_dets)
        elif tuple_item_el.tag == 'Subscript':
            dict_key_name_dets = _dict_key_name_dets_from_subscript_el(
                subscript_el=tuple_item_el, unpacking_idx=unpacking_idx)
            tuple_names_dets.append(dict_key_name_dets)
    return tuple_names_dets

def get_assigned_names(element):
    """
    Get name assignment associated with the element. The element might be the
    value or the target or something but we just want to identify the closest
    Assign ancestor and get its Name.

    Raise Exception if not found.

    If Assign appears more than once in the ancestral chain e.g.
    body-Assign-spam-eggs-Assign-targets-Name then we get a list like this:
    [body-Assign, body-Assign-spam-eggs-Assign] and we want the closest one to
    the element i.e. assign_els[-1]

    Ordered set of nodes, from parent to ancestor?
    https://stackoverflow.com/a/15645846

    :return: name type (std, dict_key, obj_attr); list of name details; and name
     as single string. For name details, a list with one item in case of std,
     two otherwise.
    :rtype: tuple
    """
    assign_els = element.xpath('ancestor-or-self::Assign')
    try:
        assign_el = assign_els[-1]
    except IndexError:
        raise IndexError(  ## in theory, guaranteed to be one given how we got the name2name_els
            f"Unable to identify ancestor Assign for element: {element}")

    std_name_dets = _std_name_dets_from_assign_el(assign_el)
    if std_name_dets:
        names_dets = [std_name_dets, ]
        return names_dets

    obj_attr_name_dets = _obj_attr_name_dets_from_assign_el(assign_el)
    if obj_attr_name_dets:
        names_dets = [obj_attr_name_dets, ]
        return names_dets

    dict_key_name_dets = _dict_key_name_dets_from_assign_el(assign_el)
    if dict_key_name_dets:
        names_dets = [dict_key_name_dets, ]
        return names_dets

    tuple_names_dets = _tuple_names_dets_from_assign_el(assign_el)
    if tuple_names_dets:
        names_dets = tuple_names_dets
        return names_dets

    return [None, ]

def get_assigned_name(element):
    names_dets = get_assigned_names(element)
    n_names = len(names_dets)
    if n_names == 0:
        raise Exception("Tried to get name but got nothing")
    elif n_names > 1:
        raise Exception(f"Tried to get name but got multiple names ({n_names})")
    else:
        name_dets = names_dets[0]
        return name_dets



import logging
from textwrap import dedent

from .. import gen_utils
from ..gen_utils import layout_comment as layout

"""
Note - plain MDV - works in some consoles where terminal output fails.

Lots in common with cli displayer but risks of DRYing probably outweigh benefits
at this stage.

Makes temp MD file and displays its content.
"""

from .. import conf

MDV_CODE_START = MDV_CODE_END = "```"

def get_message(message_dets, detail_level):
    message = dedent(message_dets.message[detail_level])
    if detail_level == conf.EXTRA:
        message = dedent(message_dets.message[conf.MAIN]) + message
    message = dedent(message)
    message = (message
        .replace(conf.PYTHON_CODE_START, '\n' + MDV_CODE_START)
        .replace('\n    ' + conf.PYTHON_CODE_END, MDV_CODE_END + '\n')
    )
    return message

def _need_snippet_displayed(overall_messages_dets, block_messages_dets, *,
        multi_block=False):
    """
    Don't need to see the code snippet displayed when it is already visible:
    * because there is only one block in snippet and there is a block message
      for it (which will display the block i.e. the entire snippet) UNLESS there
      is an overall message separating them
    Otherwise we need it displayed.
    """
    mono_block_snippet = not multi_block
    if mono_block_snippet and block_messages_dets and not overall_messages_dets:
        return False
    return True

def display(snippet, file_path, messages_dets, *,
        detail_level=conf.BRIEF,
        warnings_only=False, multi_block=False, multi_script=False):
    """
    Show by code blocks.
    """
    if warnings_only:
        options_msg = conf.WARNINGS_ONLY_MSG
    else:
        options_msg = conf.ALL_HELP_SHOWING_MSG
    intro = gen_utils.get_intro(file_path, multi_block=multi_block)
    text = [
        layout(f"""\
            # SuperHELP - Help for Humans!

            {intro}

            Currently showing {detail_level} content as requested.
            {options_msg}.

            {conf.MISSING_ADVICE_MESSAGE}

                #################################################################

                # Please give us a star on https://github.com/grantps/superhelp #

                #             so others can learn about us. Thanks!             #

                #################################################################
            """
        ),
    ]
    overall_messages_dets, block_messages_dets = messages_dets
    display_snippet = _need_snippet_displayed(
        overall_messages_dets, block_messages_dets, multi_block=multi_block)
    if display_snippet:
        line_numbered_snippet = gen_utils.get_line_numbered_snippet(snippet)
        code_desc = gen_utils.get_code_desc(file_path)
        text.append(dedent(
            f"## {code_desc}"
            f"\n{MDV_CODE_START}\n"
            + line_numbered_snippet
            + f"\n{MDV_CODE_END}"))
    for message_dets in overall_messages_dets:
        message = get_message(message_dets, detail_level)
        text.append(message)
    block_messages_dets.sort(key=lambda nt: (nt.first_line_no, nt.warning))
    prev_line_no = None
    for message_dets in block_messages_dets:
        ## display code for line number (once ;-))
        line_no = message_dets.first_line_no
        new_block = (line_no != prev_line_no)
        if new_block:
            block_has_warning_header = False
            text.append(dedent(
                f'## Code block starting line {line_no:,}'
                f"\n{MDV_CODE_START}\n"
                + message_dets.code_str
                + f"\n{MDV_CODE_END}"))
            prev_line_no = line_no
        if message_dets.warning and not block_has_warning_header:
            text.append(layout("""\
                ### Questions / Warnings

                There may be some issues with this code block you want to
                address.
                """))
            block_has_warning_header = True
        ## process message
        message = get_message(message_dets, detail_level)
        text.append(message)
    content = '\n'.join(text)
    if file_path:
        raw_file_name = gen_utils.clean_path_name(file_path)
        file_name = f'{raw_file_name}.md'
    else:
        file_name = 'superhelp.md'
    with gen_utils.make_open_tmp_file(file_name, mode='w') as tmp_dets:
        _superhelp_tmpdir, tmp_fh, fpath = tmp_dets
        tmp_fh.write(content)
    print(content)
    print(f"""\
    {'-' * 10} Content above this line saved as temp file {'-' * 10}
    Temp file: {fpath}
    """)
    if multi_script:
        input("Press any key to continue ...")



import logging

from superhelp.helpers import all_blocks_help
from ..ast_funcs import num_str_from_val, _get_var_equal_plussed, _get_var_plus_equalled
from .. import conf
from superhelp.gen_utils import layout_comment as layout

def get_num(value_el):
    """
    ## a positive number
    <Assign lineno="1" col_offset="0">
      ...
      <value>
        <Constant lineno="1" col_offset="4" type="int" value="0"/>
      </value>
    </Assign>
    
    ## a negative number
    <Assign lineno="1" col_offset="0">
      ...
      <value>
        <UnaryOp lineno="1" col_offset="4">
          <op>
            <USub/>
          </op>
          <operand>
            <Constant lineno="1" col_offset="5" type="int" value="1"/>
          </operand>
        </UnaryOp>
      </value>
    </Assign>
    """
    return num_str_from_val(value_el)

def get_var_initialised(el):
    """
    e.g. i = 0 or i = -1 or i = 1

    ## example where a positive number
    <Assign lineno="1" col_offset="0">
      <targets>
        <Name lineno="1" col_offset="0" type="str" id="counter">
          <ctx>
            <Store/>
          </ctx>
        </Name>
      </targets>
      <value>
        <Constant lineno="1" col_offset="10" type="int" value="0"/>
      </value>
    </Assign>
    """
    assign_els = el.xpath('descendant-or-self::Assign')
    if not assign_els:
        return None
    has_var_init = False
    for assign_el in assign_els:
        value_els = assign_el.xpath('value')
        if len(value_els) != 1:
            continue
        num = get_num(value_els[0])
        if num not in ['0', '-1', '1']:
            continue
        name_els = assign_el.xpath('targets/Name')
        if not name_els:
            continue
        name = name_els[0].get('id')
        has_var_init = True
        break
    var_initialised = name if has_var_init else None
    return var_initialised

def get_var_plus_equalled(el):
    """
    e.g. i += 1

    <AugAssign lineno="2" col_offset="0">
      <target>
        <Name lineno="2" col_offset="0" id="counter">
          <ctx>
            <Store/>
          </ctx>
        </Name>
      </target>
      <op>
        <Add/>
      </op>
      <value>
        <Constant lineno="2" col_offset="11" value="1"/>
      </value>
    </AugAssign>
    """
    return _get_var_plus_equalled(el)

def get_var_equal_plussed(el):
    """
    e.g. i = i + 1

    <Assign lineno="4" col_offset="0">
      <targets>
        <Name lineno="4" col_offset="0" id="counter">
          <ctx>
            <Store/>
          </ctx>
        </Name>
      </targets>
      <value>
        <BinOp lineno="4" col_offset="10">
          <left>
            <Name lineno="4" col_offset="10" id="counter">
              <ctx>
                <Load/>
              </ctx>
            </Name>
          </left>
          <op>
            <Add/>
          </op>
          <right>
            <Num lineno="4" col_offset="20" n="1"/>
          </right>
        </BinOp>
      </value>
    </Assign>
    """
    return _get_var_equal_plussed(el)

def get_var_incremented(el):
    """
    e.g. i += 1 or i = i + 1
    """
    var_incremented = get_var_plus_equalled(el)
    if not var_incremented:
        var_incremented = get_var_equal_plussed(el)
    return var_incremented

def process_el(el, vars_initialised, vars_incremented):
    """
    Look for either init or incrementing.

    Update vars_init set and, if var already in init set, in
    vars_incremented set. Once anything in vars_incremented set, return
    because we only need one. If neither, get children and process down each
    path. Then return.
    """
    var_initialised = get_var_initialised(el)
    if var_initialised:
        vars_initialised.add(var_initialised)
    var_incremented = get_var_incremented(el)
    if var_incremented:
        if var_incremented in vars_initialised:
            vars_incremented.add(var_incremented)
            return
    for child_el in el.getchildren():
        process_el(child_el, vars_initialised, vars_incremented)
    return

def get_incrementing_var(block_dets, vars_initialised, vars_incremented):
    """
    Iterate through lines in AST (depth first). Look for var init and once
    found, then incrementing. On first init-incrementing pair found, halt and
    give message about enumerate using the var name in example.
    """
    process_el(block_dets.element, vars_initialised, vars_incremented)
    if not vars_incremented:
        incrementing_var = None
    else:
        incrementing_var = vars_incremented.pop()
    return incrementing_var

@all_blocks_help()
def manual_incrementing(blocks_dets):
    """
    Look for manual handling of incrementing inside loops.
    """
    vars_initialised = set()
    vars_incremented = set()
    has_incrementing = False
    for block_dets in blocks_dets:
        incrementing_var = get_incrementing_var(
            block_dets, vars_initialised, vars_incremented)
        logging.debug(f"vars_initialised: {vars_initialised}; "
            f"vars_incremented: {vars_incremented}")
        if incrementing_var:
            has_incrementing = True
            break
    if not has_incrementing:
        return None

    summary = layout(f"""\
    ### Possible option of using `enumerate()`

    It looks like your code is manually incrementing `{incrementing_var}`. In
    Python you can use the `enumerate` function to handle this for you.
    """)
    demo = (
        layout("""\
        Here is an example of the manual approach:
        """)
        +
        layout("""\
        n = 1
        for image in images:
            if n % 10 == 0:
                print(f"Just processed image {{n}}")
            process_image(image)
            n += 1
        """, is_code=True)
        +
        layout("""\
        Here is how we can use `enumerate()` instead:
        """)
        +
        layout("""\
        for n, image in enumerate(images, 1):
            if n % 10 == 0:
                print(f"Just processed image {{n}}")
            process_image(image)
        """, is_code=True)
        +
        layout("""\

        Often you want counting from 0 in which case you don't need to specify
        the start value (0 is the default):
        """)
        +
        layout("""\
        for i, image in enumerate(images):
            ...
        """, is_code=True)
        +
        layout("""\

        You can give the enumerated value any name that makes sense but reserve
        `i` for incrementing starting at 0 and prefer `n` when starting at 1.
        """)
    )

    message = {
        conf.BRIEF: summary,
        conf.MAIN: summary + demo,
    }
    return message



from superhelp.helpers import  get_dict_comprehension_msg, \
    get_general_comprehension_msg, get_set_comprehension_msg, \
    filt_block_help
from .. import code_execution, conf
from superhelp import gen_utils
from superhelp.gen_utils import layout_comment as layout

def truncate_list(items):
    return items[: conf.MAX_ITEMS_EVALUATED]

ASSIGN_LISTCOMP_XPATH = 'descendant-or-self::Assign/value/ListComp'

@filt_block_help(xpath=ASSIGN_LISTCOMP_XPATH)
def listcomp_overview(block_dets, *, repeat=False):
    """
    Provide advice on list comprehensions and explain other types of
    comprehension available in Python.
    """
    listcomp_els = block_dets.element.xpath(ASSIGN_LISTCOMP_XPATH)
    try:
        names_items, oversized_msg = code_execution.get_collections_dets(
            listcomp_els, block_dets,
            collection_plural='lists', truncated_items_func=truncate_list)
    except Exception:
        names_items = []
        oversized_msg = ''

    plural = 's' if len(names_items) > 1 else ''
    title = layout(f"""\
    ### List comprehension{plural} used
    """)
    summary_bits = []
    for name, items in names_items:
        if items is None:
            summary_bits.append(layout(f"""
            `{name}` is a list comprehension. Unable to evaluate items.
            """))
        else:
            summary_bits.append(layout(f"""

            `{name}` is a list comprehension returning a list with
            {gen_utils.int2nice(len(items))} items: {items}
            """))
    summary = ''.join(summary_bits)
    brief_summary = summary
    main_summary = summary
    if not brief_summary:
        brief_summary = layout("""\
        List comprehension detected but unable to evaluate individual list
        comprehensions in your code.
        """)
    if not main_summary:
        if not repeat:
            main_summary = layout("""\
            Unable to evaluate individual list comprehensions in your code but
            still able to make some general comments.
            """)
        else:
            main_summary = layout("""\
            List comprehension detected but unable to evaluate individual list
            comprehensions in your code.
        """)
    if not repeat:
        other_comprehensions = (
            layout("""\
            ### Other "comprehensions"
            """)
            + get_general_comprehension_msg()
            + layout("""\

            List comprehensions aren't the only type of comprehension you can
            make. Python also lets you write Dictionary and Set Comprehensions:
            """)
            + get_dict_comprehension_msg()
            + '\n\n'
            + get_set_comprehension_msg()
            + '\n\n'
            + layout("""\

            Pro tip: don't make comprehension *in*comprehensions ;-). If it is
            hard to read it is probably better written as a looping structure.
            """)
        )
    else:
        other_comprehensions = ''

    message = {
        conf.BRIEF: title + oversized_msg + brief_summary,
        conf.MAIN: title + oversized_msg + main_summary,
        conf.EXTRA: other_comprehensions,
    }
    return message



from collections import namedtuple

from . import conf  # @UnresolvedImport
from superhelp.gen_utils import layout_comment as layout

## extract patterns e.g. from:
## /tmp/snippet.py:1:23: W291 trailing whitespace
## /tmp/snippet.py:4:1: E999 IndentationError: unexpected indent
LINT_PATTERN = fr"""^.+?:                                  ## starts with misc until first colon e.g. /tmp/snippet.py: (+? to be non-greedy)
        (?P<{conf.LINT_LINE_NO}>\d+)                       ## line_no = one or more digits after snippet.py: and before the next : e.g. 4  # @UndefinedVariable
        :\d+:\s{{1}}                                       ## one or more digits, :, and one space e.g. '1: '
        (?P<{conf.LINT_MSG_TYPE}>\w{{1}}\d{{1,3}})\s{{1}}  ## msg_type = (one letter, 1-3 digits) and one space e.g. 'E999 '
        (?P<{conf.LINT_MSG}>.*)                            ## msg = everything else e.g. 'IndentationError: unexpected indent'
    """

## https://pycodestyle.pycqa.org/en/latest/intro.html#error-codes
## https://flake8.pycqa.org/en/latest/user/error-codes.html
IGNORED_LINT_RULES = [
    'E266', 'E262',  ## I like ## before comments and # before commented out code (idea copied off Tom Eastman - thanks Tom!)
    'E305',  ## for classes I agree with 2 spaces but not functions
    'W292',  ## no newline at end of file inappropriate for snippets as opposed to modules
]

def title2msg_type(title):
    return title.upper().replace(' ', '_') + '_MSG_TYPE'
## Ensure brief AND main are the same so titles don't shift when
## changing message level
line_indentation_title = "Line indentation issues"
whitespace_title = "White space issues"
blank_line_title = "Blank line issues"
line_length_title = "Excessive line length"
unused_imports_title = "Unused imports"
undefined_names_title = "Undefined names"
## nice to keep placeholder names etc aligned with actual titles but nothing breaks if we don't
LINE_INDENTATION_MSG_TYPE = title2msg_type(line_indentation_title)
WHITESPACE_MSG_TYPE = title2msg_type(whitespace_title)
BLANK_LINES_MSG_TYPE = title2msg_type(blank_line_title)
LINE_LENGTH_MSG_TYPE = title2msg_type(line_length_title)
UNUSED_IMPORTS_MSG_TYPE = title2msg_type(unused_imports_title)
UNDEFINED_NAMES_MSG_TYPE = title2msg_type(undefined_names_title)

def consolidated_msg_type(msg_type):
    if msg_type.startswith('E1'):
        msg_type = LINE_INDENTATION_MSG_TYPE
    elif msg_type.startswith('E2'):
        msg_type = WHITESPACE_MSG_TYPE
    elif msg_type.startswith('E3'):
        msg_type = BLANK_LINES_MSG_TYPE
    elif msg_type == 'E501':
        msg_type = LINE_LENGTH_MSG_TYPE
    elif msg_type == 'F401':
        msg_type = UNUSED_IMPORTS_MSG_TYPE
    elif msg_type == 'F821':
        msg_type = UNDEFINED_NAMES_MSG_TYPE
    return msg_type

LevelMsgs = namedtuple('LintMsgsByLevel', 'brief, main, extra')

undefined_names_level_msgs = LevelMsgs(
    layout(f"""

        #### {undefined_names_title}

        The linter has raised questions about undefined names.

        """),
    layout(f"""\

        #### {undefined_names_title}

        The linter has raised questions about undefined names. That isn't a
        problem if the snippet was extracted from a larger piece of code and the
        names were defined before the snippet.

        """),
    ''
)

unused_imports_level_msgs = LevelMsgs(
    layout(f"""\

        #### {unused_imports_title}

        One or more imports have not been used in the snippet.

        """),
    layout(f"""\

        #### {unused_imports_title}

        One or more imports have not been used in the snippet. That isn't a
        problem if the snippet was extracted from a larger piece of code and the
        imports are used later after the snippet.

        """),
    ''
)

line_length_level_msgs = LevelMsgs(
    layout(f"""\

        #### {line_length_title}

        One or more lines are longer than the recommended 79 characters. This is
        not necessarily a problem but long lines should be an exception to the
        rule.

        """),
    layout(f"""\

        #### {line_length_title}

        One or more lines are longer than the recommended 79 characters. This is
        not necessarily a problem given that we have wider monitors than when
        the guidelines were formulated. But long lines should be an exception to
        the rule. All being equal, short lines are easier to read and understand
        than long lines. There are multiple strategies for shortening lines but
        the overall goal has to be readability. Sometimes we have to live with
        broken "rules". And that's official. Read PEP 8 - the official Python
        style guide - especially the section "A Foolish Consistency is the
        Hobgoblin of Little Minds".

        """),
    ''
)

blank_lines_level_msgs = LevelMsgs(
    layout(f"""

        #### {blank_line_title}

        The linter has raised questions about blank lines.

        """),
    layout(f"""

        #### {blank_line_title}

        The linter has raised questions about blank lines. Class definitions
        should have two blank lines before. On function definitions there is
        more flexibility. It should either be one or two.

        """),
    ''
)

whitespace_level_msgs = LevelMsgs(
    layout(f"""

        #### {whitespace_title}

        The linter has raised questions about "whitespace" (tabs, spaces).

        """),
    layout(f"""

        #### {whitespace_title}

        The linter has raised questions about "whitespace" (tabs, spaces). Even
        when whitespace doesn't seem to matter it is best to follow Python
        whitespace conventions when writing Python. Conventions may differ in
        other languages.

        """),
    ''
)

line_indentations_level_msgs = LevelMsgs(
    layout(f"""

        #### {line_indentation_title}

        The linter has raised questions about indentation. There are at least
        two styles of indentation. Whichever you follow be consistent.
    """),
    (
        layout(f"""\

            #### {line_indentation_title}

            The linter has raised questions about indentation. There are at
            least two styles of indentation:

            1) Visual line continuation e.g. note how `parameter_3` lines up
            with the opening parenthesis:

            """)
        +
        layout("""\

            def function_with_long_name(parameter_1, parameter_2,
                                        parameter_3, parameter_4):
                '''
                Doc string for function
                '''
            """, is_code=True)
        +
        layout("""\
            Pros:

            * a common convention

            Cons:

            * every time you rename a function you have to realign parameters

            * code will generally not align tidily with multiples of standard
            indentation

            * room for parameters squeezed

            * you have to look to the right to make sense of the function

            2) Leftwards alignment with multiples of standard indentation e.g.

            """)
        +
        layout("""\

            def function_with_long_name(parameter_1, parameter_2,
                    parameter_3, parameter_4):
                '''
                Doc string for function
                '''

            """, is_code=True)
        +
        layout("""\

            or

            """)
        +
        layout("""\

            def function_with_long_name(
                    parameter_1, parameter_2, parameter_3, parameter_4):
                '''
                Doc string for function
                '''
            """, is_code=True)
        +
        layout("""\

            or

            """)
        +
        layout("""\

            def function_with_long_name(
                    parameter_1,
                    parameter_2,
                    parameter_3,
                    parameter_4):
                '''
                Doc string for function
                '''

            """, is_code=True)
        +
        layout("""\

            Pros:

            * readability is enhanced because the main content of the code is as
            leftward as possible

            * beautiful alignment across entire module

            Cons:

            * you may need to change your IDE's default indenting behaviour

            Whichever of the two style you follow, be consistent.
            """)
        ),
    layout("""\

        See <https://rhodesmill.org/brandon/slides/2012-11-pyconca/#id183> for
        some thought-provoking ideas on code style and much more.

        """)
)

CUSTOM_LINT_MSGS = {
    UNDEFINED_NAMES_MSG_TYPE: undefined_names_level_msgs,
    LINE_LENGTH_MSG_TYPE: line_length_level_msgs,
    UNUSED_IMPORTS_MSG_TYPE: unused_imports_level_msgs,
    BLANK_LINES_MSG_TYPE: blank_lines_level_msgs,
    WHITESPACE_MSG_TYPE: whitespace_level_msgs,
    LINE_INDENTATION_MSG_TYPE: line_indentations_level_msgs,
}



from collections import namedtuple

from superhelp.helpers import all_blocks_help
from .. import ast_funcs, conf
from superhelp.gen_utils import layout_comment as layout

NTDets = namedtuple('NamedTupleDetails', 'name, label, fields_str, fields_list')

def get_named_tuple_dets(named_tuple_el):
    """
    Get name, label, fields.
    """
    assign_block_el = named_tuple_el.xpath('ancestor-or-self::Assign')[-1]
    name = assign_block_el.xpath('targets/Name')[0].get('id')
    label, fields_list = ast_funcs.get_nt_lbl_flds(assign_block_el)
    fields_str = ', '.join(fields_list)
    named_tuple_dets = NTDets(name, label, fields_str, fields_list)
    return named_tuple_dets

def get_named_tuples_dets(blocks_dets):
    all_named_tuples_dets = []
    for block_dets in blocks_dets:
        func_name_els = block_dets.element.xpath(
            'descendant-or-self::value/Call/func/Name')
        if not func_name_els:
            continue
        named_tuple_els = [func_name_el for func_name_el in func_name_els
            if func_name_el.get('id') == 'namedtuple']
        named_tuples_dets = [
            get_named_tuple_dets(named_tuple_el)
            for named_tuple_el in named_tuple_els]
        all_named_tuples_dets.extend(named_tuples_dets)
    return all_named_tuples_dets

@all_blocks_help()
def named_tuple_overview(blocks_dets):
    """
    Look for named tuples and explain how they can be enhanced.
    """
    named_tuples_dets = get_named_tuples_dets(blocks_dets)
    if not named_tuples_dets:
        return None

    example_dets = named_tuples_dets[0]
    first_field = example_dets.fields_list[0]
    enhancement = (
            layout("""\
            ### Named Tuple Enhancements

            Named tuples can be enhanced to make them even more useful -
            especially when debugging. The label can be expanded beyond the
            variable name; and the entire named tuple or individual fields can
            be given their own doc strings.

            For example:
            """)
            +
            layout(f"""\
            {example_dets.name} = namedtuple("{example_dets.label}",
                "{example_dets.fields_str}")
            {example_dets.name}.__doc__ += "\\n\\nExtra comments"
            {example_dets.name}.{first_field}.__doc__ = "Specific comment for {first_field}"
            ## etc
            """, is_code=True)
        )
    defaults = (
        layout("""\

        Default arguments are another nice option (added in Python 3.7). For
        example the following named tuple has a default IQ of 100:
        """)
        +
        layout("""\
        People = namedtuple('PeopleDets', 'name, IQ', defaults=(100, ))
        """, is_code=True)
        +
        layout("""\
        The official documentation has more details.
        """)
    )

    message = {
        conf.BRIEF: enhancement,
        conf.MAIN: enhancement + defaults,
    }
    return message
    


"""
Method helpers are effectively function helpers and are covered there.
"""
from collections import defaultdict

from superhelp.helpers import filt_block_help
from .. import conf
from superhelp.gen_utils import get_nice_str_list, layout_comment as layout

CLASS_XPATH = ('descendant-or-self::ClassDef')

@filt_block_help(xpath=CLASS_XPATH, warning=True)
def getters_setters(block_dets, *, repeat=False):
    """
    Look for getters and setters and suggest @property if appropriate.
    """
    class_els = block_dets.element.xpath(CLASS_XPATH)
    if not class_els:
        return None
    class_getter_setter_methods = defaultdict(list)
    for class_el in class_els:
        class_name = class_el.get('name')
        method_els = class_el.xpath('body/FunctionDef')
        for method_el in method_els:
            method_name = method_el.get('name')
            if method_name.startswith(('set_', 'get_')):
                class_getter_setter_methods[class_name].append(method_name)
    if not class_getter_setter_methods:
        return None

    title = layout("""\
    ### Alternative to getters and setters
    """)
    simple_class_msg_bits = []
    for class_name, method_names in sorted(class_getter_setter_methods.items()):
        multiple = len(method_names) > 1
        if multiple:
            nice_list = get_nice_str_list(method_names, quoter='`')
            simple_class_msg_bits.append(layout(f"""\

            Class `{class_name}` has the following methods that look like either
            getters or setters: {nice_list}.
            """))
        else:
            method_type = (
                'getter' if method_name.startswith('get_') else 'setter')
            simple_class_msg_bits.append(layout(f"""\

            Class `{class_name}` has a `{method_names[0]}` method that looks
            like a {method_type}.
            """))
    simple_class_msg = ''.join(simple_class_msg_bits)
    if not repeat:
        properties_option = layout("""\

        Python doesn't need getters and setters. Instead, you can use
        properties. These are easily added using decorators e.g. `@property`.
        """)
        tm = '\N{TRADE MARK SIGN}'
        why_getters_etc = layout(f"""\

        A good discussion of getters, setters, and properties can be found at
        <https://www.python-course.eu/python3_properties.php>.

        Getters and setters are usually added in other languages such as Java
        because direct attribute access doesn't give the ability to calculate
        results or otherwise run a process when a value is accessed / written.

        And it is common for lots of getters and setters to be added, whether or
        not they are actually needed - Just In Case{tm}. The fear is that if you
        point other code to an attribute, and you later need to process the
        attribute or derive it before it is served up or stored, then you'll
        need to make a breaking change to your code. All the client code
        referencing the attribute will have to be rewritten to replace direct
        access with a reference to the appropriate getter or setter.
        Understandably then the inclination is to point to a getter or setter in
        the first case even if it doesn't actually do anything different (for
        now at least) from direct access. Using these getters and setters is
        wasteful, and bloats code unnecessarily, but it avoids the worse evil of
        regularly broken interfaces. The benefit is that you can change
        implementation later if you need to and nothing will break. But in
        Python there is a much better way :-).
        """)
        comparison = (
            layout(f"""\

            Let's compare the getter / setter approach and the property
            approach.

            #### Using getters / setters
            """)
            +
            layout("""
            class Person:

                def __init__(self, name):
                    self.__name = name

                def get_name(self):
                    return self.__name

                def set_name(self, name):
                    if name is not None:
                        self.__name = name
            """, is_code=True)
            +
            layout("""\
            #### Using properties
            """)
            +
            layout("""
            class Person:

                def __init__(self, name):
                    self.name = name

                @property  ## the getter
                def name(self):
                    return self.__name

                @name.setter  ## the setter
                def name(self, name):
                    if name is not None:
                        self.__name = name
            """, is_code=True)
            +
            layout("""\

            We must define the getter earlier in the script than the setter
            because the setter references the getter name. If it is not already
            defined above it we will get a `NameError` because Python doesn't
            yet know the variable_name part of the `@<variable_name>.setter`.

            In the `\_\_init\_\_` method we can either directly set the
            "private" variable `self.__name` (note - unenforced in Python)
            or set the public variable `self.name` and pass through the
            setter. Passing through the setter makes more sense - presumably
            there are some smarts in the setter we want applied otherwise we
            wouldn't have gone to the trouble of making it ;-).

            You will have also noticed that the method name is defined twice
            without the second one overwriting the first (the standard
            behaviour). The decorators take care of all that. The only thing to
            remember is to use exactly the same attribute name in three places
            (assuming both getting and setting):
            """)
            +
            layout("""\
            @property
            def here():  ## <== 1
                ...

            @here.setter  ## <== 2
            def here():  ## <== 3
                ...
            """, is_code=True)
        )
        deleter = (
            layout("""\

            Python also has a `deleter` decorator which handle deletion of the
            attribute e.g.
            """)
            +
            layout("""\
            @name.deleter
            def name(self):
                ...
            """, is_code=True)
        )
    else:
        properties_option = ''
        why_getters_etc = ''
        comparison = ''
        deleter = ''

    brief_msg = title + simple_class_msg + properties_option
    main_msg = (title + simple_class_msg + properties_option
        + why_getters_etc + comparison)
    extra_msg = deleter
    message = {
        conf.BRIEF: brief_msg,
        conf.MAIN: main_msg,
        conf.EXTRA: extra_msg,
    }
    return message

@filt_block_help(xpath=CLASS_XPATH, warning=True)
def selfless_methods(block_dets, *, repeat=False):
    """
    Look for class methods that don't use self as candidates for @staticmethod
    decorator. Note - not worried about detecting sophisticated cases with
    packing etc although it doesn't have to be named "self".
    """
    class_els = block_dets.element.xpath(CLASS_XPATH)
    if not class_els:
        return None
    class_selfless_methods = defaultdict(list)
    for class_el in class_els:
        class_name = class_el.get('name')
        method_els = class_el.xpath('body/FunctionDef')
        for method_el in method_els:
            method_name = method_el.get('name')
            arg_els = method_el.xpath('args/arguments/args/arg')
            if not arg_els:
                continue
            first_arg_name = arg_els[0].get('arg')
            used_first_names = [
                name_el for name_el in method_el.xpath('descendant::Name')
                if name_el.get('id') == first_arg_name]
            if not used_first_names:
                class_selfless_methods[class_name].append(method_name)
    if not class_selfless_methods:
        return None

    title = layout("""\
    ### Method doesn't use instance
    """)
    simple_class_msg_bits = []
    for class_name, method_names in sorted(class_selfless_methods.items()):
        multiple = len(method_names) > 1
        if multiple:
            nice_list = get_nice_str_list(method_names, quoter='`')
            simple_class_msg_bits.append(layout(f"""\

            Class `{class_name}` has the following methods that don't use the
            instance (usually called `self`): {nice_list}.
            """))
        else:
            simple_class_msg_bits.append(layout(f"""\

            Class `{class_name}` has a `{method_names[0]}` method that doesn't
            use the instance (usually called `self`).
            """))
    simple_class_msg = ''.join(simple_class_msg_bits)
    if not repeat:
        staticmethod_msg = layout("""\

        If a method doesn't use the instance it can be either pulled into a
        function outside the class definition or decorated with `@staticmethod`.
        The decorator stops the method expecting the instance object to be
        supplied as the first argument.
        """)
        staticmethod_demo = (
            layout("""
            For example, instead of:
            """)
            +
            layout("""
            def calc_approx_days(self, years):  ## <== self not used!
                return round(years * 365.25)
            """, is_code=True)
            +
            layout("""
            you could write:
            """)
            +
            layout("""
            @staticmethod
            def calc_approx_days(years):
                return round(years * 365.25)
            """, is_code=True)
        )
        call_it_self = layout("""\

        It is not obligatory to call the first parameter of a bound method
        `self` but you should call it that unless you have a good reason to
        break convention.
        """)
    else:
        staticmethod_msg = ''
        staticmethod_demo = ''
        call_it_self = ''

    brief_msg = title + simple_class_msg + staticmethod_msg
    main_msg = title + simple_class_msg + staticmethod_msg + staticmethod_demo
    extra_msg = call_it_self
    message = {
        conf.BRIEF: brief_msg,
        conf.MAIN: main_msg,
        conf.EXTRA: extra_msg,
    }
    return message

@filt_block_help(xpath=CLASS_XPATH, warning=True)
def one_method_classes(block_dets, *, repeat=False):
    """
    Look for classes with only one method (other than __init__) and suggest a
    simple function as an alternative.
    """
    class_els = block_dets.element.xpath(CLASS_XPATH)
    if not class_els:
        return None
    classes_sole_methods = []
    for class_el in class_els:
        method_els = class_el.xpath('body/FunctionDef')
        method_names = [method_el.get('name') for method_el in method_els]
        non_init_method_names = [method_name for method_name in method_names
            if method_name != '__init__']
        n_non_init_methods = len(non_init_method_names)
        if n_non_init_methods < 2:
            class_name = class_el.get('name')
            try:
                sole_method_name = non_init_method_names.pop()
            except IndexError:
                sole_method_name = None
            classes_sole_methods.append((class_name, sole_method_name))
    if not classes_sole_methods:
        return None

    multi_sole = len(classes_sole_methods) > 1
    class_plural = 'es' if multi_sole else ''
    class_have_has = 'have' if multi_sole else 'has'
    func_plural = 's' if multi_sole else ''
    summary = layout(f"""\

    ### Possible option of converting class{class_plural} to single
    function{func_plural}

    The following class{class_plural} only {class_have_has} one main method at
    most (excluding `__init__`):
    """)
    n_methods_msg_bits = []
    for class_name, method_name in classes_sole_methods:
        method2use = (f"`{method_name}`" if method_name
            else 'nothing but `__init__` (if even that)')
        n_methods_msg_bits.append(layout(f"""\
        - `{class_name}`: {method2use}
        """))
    n_methods_msg = ''.join(n_methods_msg_bits)
    if not repeat:
        not_just_oo = layout("""\

        Python allows procedural, object-oriented, and functional styles of
        programming. Event-based programming is also used in GUI contexts, for
        example. Programmers coming to Python from languages that only support
        object-orientation sometimes overdo the classes when there is a simpler,
        more elegant way of writing readable code in Python.

        If only a simple function is required, then write a simple function.

        Note - there may be exceptions. It has been suggested that the class
        structure can make it easier to test intermediate state rather than just
        function outputs. So, as with most things, it depends.
        """)
        function_option = layout(f"""\

        It may be simpler to replace the class{class_plural} with simple
        functions.
        """)
        function_demo = (
            layout('''\
            class GetSquare:

                def __init__(self, num):
                    self.num = num

                def get_square(self):
                    """
                    Get square of number
                    """
                    return self.num ** 2
            ''', is_code=True)
            +
            layout("""\
            with:
            """)
            +
            layout('''\
            def get_square(num):
                """
                Get square of number ...
                """
                return self.num ** 2
            ''', is_code=True)
            +
            layout("""\

            Sometimes, less is more. Bugs need somewhere to hide and
            unnecessarily verbose code gives them plenty of room to hide.
            """)
        )
    else:
        not_just_oo = ''
        function_option = ''
        function_demo = ''

    message = {
        conf.BRIEF: summary + n_methods_msg + function_option,
        conf.MAIN: summary + n_methods_msg + function_option + function_demo,
        conf.EXTRA: not_just_oo,
    }
    return message



import logging

from . import conf, name_utils
from superhelp.gen_utils import get_nice_str_list, layout_comment as layout

def get_val(pre_block_code_str, block_code_str,
        name_type, name_details, name_str):
    """
    Executing supplied code from end users - nope - nothing to see here from a
    security point of view ;-) Needs addressing if this code is ever used as a
    service for other users.

    Note - can be the source of mysterious output in stdout (e.g. exec a print
    function).

    :param str name_type: e.g. conf.STD_NAME. Lets us know how to handle the
     parts of the name e.g. dict name and key name.
    :param list name_details: e.g. name; dict and key; obj and attr
    :param str name_str: name as string e.h. Family.pet or capitals['NZ']
    :return: value if possible. Raises KeyError if unable to get value.
    """
    exp_dets = {}
    try:
        exec(pre_block_code_str + block_code_str, exp_dets)
    except ImportError as e:
        logging.debug(
            f"Import problem running {__file__} (specifically {__name__}): {e}")
        raise ImportError("SuperHELP only has modules from the Python standard "
            "library installed - it looks like your snippet relies on a module "
            "from outside the standard library.")
    if name_type == conf.STD_NAME:
        val = exp_dets[name_details[0]]
    elif name_type == conf.OBJ_ATTR_NAME:
        obj_name, attr_name = name_details
        try:
            val = getattr(exp_dets[obj_name], attr_name)
        except AttributeError:
            raise KeyError(f"Unable to find name '{name_str}' "
                f"in code_str\n{block_code_str}")
    elif name_type == conf.DICT_KEY_NAME:
        dict_name, key_name = name_details
        try:
            val = exp_dets[dict_name][key_name]
        except NameError:
            raise KeyError(f"Unable to find name '{name_str}' "
                f"in code_str\n{block_code_str}")
    else:
        raise Exception(f"Unexpected name_type: '{name_type}'")
    return val

def get_collections_dets(named_els, block_dets, *,
        collection_plural, truncated_items_func):
    """
    Get information on collections - names with associated items, plus a string
    message (empty str if no oversized items) which can be assembled as part of
    a full helper message. If an item cannot be evaluated, it is set to None.

    :return: names_items: list of (name, items) tuples, and oversized_msg (str)
    :rtype: tuple
    """
    names_items = []
    oversized_names = []
    for named_el in named_els:
        names_dets = name_utils.get_assigned_names(named_el)
        for name_dets in names_dets:
            try:
                items = get_val(
                    block_dets.pre_block_code_str, block_dets.block_code_str,
                    name_dets.name_type, name_dets.name_details,
                    name_dets.name_str)
            except KeyError:
                items = None
            else:
                if len(items) > conf.MAX_ITEMS_EVALUATED:
                    items = truncated_items_func(items)
                    oversized_names.append(name_dets.name_str)
            names_items.append((name_dets.name_str, items))
    if oversized_names:
        multi_oversized = len(oversized_names) > 1
        if multi_oversized:
            nice_names = get_nice_str_list(oversized_names, quoter='`')
            oversized_msg = layout(f"""\

            Because the following {collection_plural} were large SuperHELP has
            only examined the first {conf.MAX_ITEMS_EVALUATED} items:
            {nice_names}
            """)
        else:
            oversized_msg = layout(f"""\

            Because `{name_dets.name_str}` is large SuperHELP has only examined
            the first {conf.MAX_ITEMS_EVALUATED} items.
            """)
    else:
        oversized_msg = ''
    return names_items, oversized_msg




from . import cli_conf

DEFAULT_ANSI_COLOUR_BYTE_STR = '\033[0m'

## ANSI colour codes from http://www.lihaoyi.com/post/BuildyourownCommandLinewithANSIescapecodes.html
dark = {
    'H1_COLOUR': 231,
    'H2_COLOUR': 153,
    'H3_COLOUR': 117,
    'H4_COLOUR': 109,
    'H5_COLOUR': 65,
    'ERROR': 124,  ## red
    'LOW_VIS_COLOUR': 59,  ## low visibility
    'TEXT': 188,
    'TL': 59,
    'CODE_COLOUR': 102,  ## code
    'OPERATOR_WORD': 89,  ## Magenta
    'KEYWORD': 28,  ## Green
    'QUOTE_MARKS': 196,  ## Rust
    'OPERATOR': 231,  ## white
    'PUNCTUATION': 231,  ## white
    'INTEGER': 231,  ## white
    'FLOAT': 231,  ## white
}
light = {
    'H1_COLOUR': 17,
    'H2_COLOUR': 21,
    'H3_COLOUR': 24,
    'H4_COLOUR': 235,
    'H5_COLOUR': 52,
    'ERROR': 124,  ## red
    'LOW_VIS_COLOUR': 59,  ## low visibility
    'TEXT': 233,
    'TL': 59,
    'CODE_COLOUR': 236,  ## code
    'OPERATOR_WORD': 89,  ## Magenta
    'KEYWORD': 28,  ## Green
    'QUOTE_MARKS': 196,  ## Rust
    'OPERATOR': 232,  ## black
    'PUNCTUATION': 232,  ## black
    'INTEGER': 232,  ## black
    'FLOAT': 232,  ## black
}

def set_global_colours(theme_name):
    """
    Given I had no desire to completely refactor the CLI code I focused the evil
    in this one place. If you read this, please forgive me.
    """
    theme = globals()[theme_name]
    global H1_COLOUR
    global TOKEN_NAME_TO_HL_COLOUR
    global LEVEL2COLOUR
    global BOUNDS2COLOUR
    global TEXT
    global CODE_COLOUR
    global LOW_VIS_COLOUR
    TOKEN_NAME_TO_HL_COLOUR = {
        "Comment": theme['LOW_VIS_COLOUR'],
        "Error": theme['ERROR'],
        "Generic": theme['H2_COLOUR'],
        "Keyword": theme['KEYWORD'],
        "Name": theme['H1_COLOUR'],
        "Number": theme['H4_COLOUR'],
        "Operator": theme['OPERATOR'],
        "String": theme['H4_COLOUR'],
        "Operator.Word": theme['OPERATOR_WORD'],
        "Literal.String.Single": theme['QUOTE_MARKS'],
        "Literal.String.Double": theme['QUOTE_MARKS'],  ## double-quoted string
        "Punctuation": theme['PUNCTUATION'],
        "Literal.Number.Integer": theme['INTEGER'],
        "Literal.Number.Float": theme['FLOAT'],
        "Keyword.Constant": theme['KEYWORD'],
    }
    LEVEL2COLOUR = {
        1: theme['H1_COLOUR'],
        2: theme['H2_COLOUR'],
        3: theme['H3_COLOUR'],
        4: theme['H4_COLOUR'],
        5: theme['H5_COLOUR'],
    }
    BOUNDS2COLOUR = {
        cli_conf.CODE_BOUNDS: theme['H2_COLOUR'],
        cli_conf.STRONG_BOUNDS: theme['H2_COLOUR'],
        cli_conf.LINK_BOUNDS: theme['H2_COLOUR'],
        cli_conf.EMPH_BOUNDS: theme['H3_COLOUR'],
    }
    TEXT = theme['TEXT']
    H1_COLOUR = theme['H1_COLOUR']
    CODE_COLOUR = theme['CODE_COLOUR']
    LOW_VIS_COLOUR = theme['LOW_VIS_COLOUR']

ADMONS = { ## Will be extended dynamically to avoid pointless recalculation if admonitions with other keys found in markdown text received
    'note': 'H3_COLOUR',
    'warning': 'ERROR',
    'attention': 'H1_COLOUR',
    'hint': 'H4_COLOUR',
    'summary': 'H1_COLOUR',
    'hint': 'H4_COLOUR',
    'question': 'H5_COLOUR',
    'danger': 'ERROR',
    'dev': 'H5_COLOUR',
    'hint': 'H4_COLOUR',
    'caution': 'H2_COLOUR',
}

def colourise(text, colour, *, reverse=False, bold=False, no_reset=False):
    """
    Colour is toggled on and will keep being applied unless toggled off (back to
    default colour).

    We apply this colour to everything except for nested content. That gets its
    own colours depending on what type of thing it is e.g. code.

    The only reason you'd supply an empty string to this is if you were setting
    a colour and had no_reset=True. Otherwise an empty line only would be
    coloured which makes no sense.

    Testing: print('Should be red spam', colourise('spam', cli_conf.ERROR))

    :param str text: text (if any) to be coloured.
    :param int colour: the colour to be applied to this text and, if no_reset
     is True, all subsequent text until toggled off (if ever).
    :param bool no_reset: colour set kept toggled on so will colour subsequent
     content until toggled off.
    """
    reset_colour = '' if no_reset else DEFAULT_ANSI_COLOUR_BYTE_STR
    for (start, end), inner_colour in BOUNDS2COLOUR.items():
        if start in text:
            if start == cli_conf.LINK_START:
                uon, uoff = "\033[4m", "\033[24m"
            else:
                uon, uoff = '', ''
            text = text.replace(
                start, colourise('', inner_colour, no_reset=True) + uon)
            text = text.replace(
                end, uoff + colourise('', colour, no_reset=True))
    ansi_reverse = cli_conf.REVERSED if reverse else ''
    ansi_bold = cli_conf.BOLD if bold else ''
    text = f"\033[38;5;{colour}m{ansi_bold}{ansi_reverse}{text}{reset_colour}"
    return text

def colourise_low_vis(text):
    return colourise(text, LOW_VIS_COLOUR)

def colourise_plain(text, **_kwargs):
    """
    Useful when a tag is not found.

    Used in context where multiple args supplied but we ignore them here.
    """
    return colourise(text, TEXT)



from superhelp.helpers import filt_block_help
from .. import conf
from superhelp.helpers import get_aop_msg
from superhelp.gen_utils import layout_comment as layout

def get_open_cm_msg():
    return (
        layout("""\

        For example, every time we open a file we need to make sure it is closed
        when we are finished. It is not that hard to handle that yourself in
        simple cases but can become much harder - perhaps the code exits in
        multiple places (perhaps as a result of an `exception`). You have to
        remember to handle the tidy-up in every place an exit might occur.
        Either you will forget one of them or you'll be repeating code all over
        the place violating the DRY principle - namely:

        > "Don't Repeat Yourself" (The Pragmatic Programmer)

        WET presumably stands for "Write Every Time", "Write Everything Twice",
        "We Enjoy Typing" or "Waste Everyone's Time" ;-)

        If you don't use a context manager, problems can emerge as a snippet
        gradually evolves. And it usually all starts so innocently :-). Here's
        one example of the problem unfolding:
        """)
        +
        layout("""\
        ## 1) Not using context manager (with) not really a problem
        f = open(fname)
        text = f.read()
        f.close()  # <------------------ First one (I hope this is the only one)

        ## 2) Maybe we need to handle empty files
        f = open(fname)
        text = f.read()
        if not text:
            raise Exception(f"{fname} is empty")  # <------ errr ... is the file
        f.close()                                 # handle going to get closed?

        ## 3) We remember to close the file handle if an exception
        f = open(fname)
        text = f.read()
        if not text:
            f.close()  # <---------------------------- Rats! I need a second one
            raise Exception(f"{fname} is empty")
        f.close()

        ## 4) We want the index of TAG for some reason
        f = open(fname)
        text = f.read()
        if not text:
            f.close()
            raise Exception(f"{fname} is empty")
        idx = text.index(TAG) # <--------------- Hmmm - will raise ValueError if
        f.close()             #                 TAG not in text? File left open?

        ## 5) We remember to handle the exception
        f = open(fname)
        text = f.read()
        if not text:
            f.close()
            raise Exception(f"{fname} is empty")
        try:
            idx = text.index(TAG)
        except ValueError:
            f.close()  # <----------------- Three already! I'm lucky it's only a
            raise      #                    one-liner each time.
        f.close()
        """, is_code=True)
        +
        layout("""\

        The code is much cleaner and robust using a simple context manager :-)
        """)
        +
        layout("""\
        with open(fname) as f:
            text = f.read()
            if not text:
                raise Exception(f"{fname} is empty")
            idx = text.index(TAG)
        ## If my code gets here, i.e. past the indented block inside the context
        ## manager,  we are guaranteed to have freed up the file - nice!
        """, is_code=True)
    )

WITH_XPATH = 'descendant-or-self::With'

def with_is_using_open(with_el):
    func_name_els = with_el.xpath(
        'descendant::items/withitem/context_expr/Call/func/Name')
    if len(func_name_els) != 1:
        return False
    func_name_el = func_name_els[0]
    func_name = func_name_el.get('id')
    return func_name == 'open'

@filt_block_help(xpath=WITH_XPATH)
def content_manager_overview(block_dets, *, repeat=False):
    """
    Explain context managers.
    """
    with_els = block_dets.element.xpath(WITH_XPATH)
    using_open_cm = any([with_is_using_open(with_el) for with_el in with_els])
    if not with_els:
        return None

    title = layout("""\
    ### Context manager(s) used
    """)
    if using_open_cm:
        summary = layout("""\

        Your code includes the commonly used file opening context manager.
        """)
    else:
        summary = layout("""\

        Your code uses a context manager.
        """)
    brief_usage = layout("""\

        Context managers take care of anything required when we enter a block of
        code or exit it.
        """)
    if not repeat:
        brief_example = layout("""\

        For example, every time we open a file we need to make sure it is closed
        when we are finished. Or when we finish with database connections and
        cursors we need to clean up after ourselves. There may be some things we
        need to do when we start a code block as well.
        """)
        long_example = get_open_cm_msg()
        aop = get_aop_msg()
    else:
        brief_example = ''
        long_example = ''
        aop = ''

    message = {
        conf.BRIEF: title + summary + brief_usage + brief_example,
        conf.MAIN: title + summary + brief_usage + long_example,
        conf.EXTRA: aop,
    }
    return message

def has_with_ancestor(open_el):
    with_els = open_el.xpath('ancestor::With')
    if not with_els:
        return False
    with_el = with_els[-1]
    func_name_els = with_el.xpath('items/withitem/context_expr/Call/func/Name')
    if len(func_name_els) != 1:
        return False
    func_name_el_under_with = func_name_els[0]
    return open_el == func_name_el_under_with

FUNC_NAME_XPATH = 'descendant-or-self::Call/func/Name'

@filt_block_help(xpath=FUNC_NAME_XPATH, warning=True)
def file_cm_needed(block_dets, *, repeat=False):
    """
    Look for opening of file without a context managers - recommend use of the
    "with open" context manager.
    """
    func_name_els = block_dets.element.xpath(FUNC_NAME_XPATH)
    if not func_name_els:
        return None
    open_els = []
    for func_name_el in func_name_els:
        func_name = func_name_el.get('id')
        if func_name == 'open':
            open_els.append(func_name_el)
    if not open_els:
        return None
    missing_cm = not all([has_with_ancestor(open_el) for open_el in open_els])
    if not missing_cm:
        return None

    title = layout("""\
    ### File opened without context manager
    """)
    summary = layout("""
    Your code opens a file without using a context manager.
    """)
    if not repeat:
        reasons = layout("""\

        Using a context manager is easy (actually, writing them isn't even that
        hard) but using the standard ones has big advantages.
        """)
        long_example = get_open_cm_msg()
        aop = get_aop_msg()
    else:
        reasons = ''
        long_example = ''
        aop = ''

    message = {
        conf.BRIEF: title + summary,
        conf.MAIN: title + summary + reasons + long_example,
        conf.EXTRA: aop,
    }
    return message



from datetime import datetime, timedelta
import json
import logging
import os

import requests

def pypi_stats():
    url = 'https://pypistats.org/api/packages/superhelp/overall'
    r = requests.get(url)
    res = r.json()
    downloads = [
        (d['date'], d['downloads']) for d in res['data']
        if d['category'] == 'without_mirrors']
    tot_downloads = sum(int(download_n) for _date, download_n in downloads)
    return ("SuperHELP - Total PyPI downloads (non-mirrored): "
        f"{tot_downloads:,}")

def github_stats():
    """
    Better to look at Insights>traffic (available to me given I have upload
    permissions). But only 14 days. And what python-trending cares about is
    stars - that's it!
    """
    url = 'https://api.github.com/repos/grantps/superhelp/releases'
    #url = 'https://api.github.com/repos/psf/requests/releases'
    headers = {'Accept': 'application/vnd.github.v3+json'}
    r = requests.get(url, headers=headers)
    if r.status_code != 200:
        print("Unable to get data from GitHub :-(")
        return
    data = r.json()
    tot_downloads = 0
    for release in data:
        for asset in release.get('assets', []):
            tot_downloads += int(asset['download_count'])
    return f"SuperHELP - Total GitHub Asset Downloads: {tot_downloads:,}"

def binderhub_stats():
    yesterday = datetime.today().date() - timedelta(days=1)
    yesterday_str = yesterday.isoformat()
    url = f"https://archive.analytics.mybinder.org/events-{yesterday_str}.jsonl"
    r = requests.get(url)
    if r.status_code != 200:
        raise Exception(f"Unable to read data - HTTP status code {r.status_code}")
    raw_jsonl = r.text
    lines = [line for line in raw_jsonl.split('\n') if line]
    records = []
    for n, line in enumerate(lines, 1):
        try:
            line_data = json.loads(line)
            spec = line_data['spec'].split('/')[0]
            if 'superhelp' not in spec.lower():
                continue
            records.append(line_data)
        except json.JSONDecodeError as e:
            logging.debug(f"Failed on line {n:,} out of {len(lines)} total lines "
                f"with {e} - original line '{line}'")
    for record in records:
        print(record)
    return f"{len(records):,} total records on '{yesterday_str}'"


print(pypi_stats())
#print(binderhub_stats())
#print(github_stats())




import logging
import re
import textwrap

from markdown.util import etree  # @UnresolvedImport
from pygments import lex, token  # @UnresolvedImport
from pygments.lexers import get_lexer_by_name  # @UnresolvedImport

from . import cli_colour, cli_conf

## monkey patch so invisible non-text is included in wrapping calculations making a mess of it
def _wrap_text_chunks_only(self, chunks):
    """_wrap_chunks(chunks : [string]) -> [string]

    Wrap a sequence of text chunks and return a list of lines of
    length 'self.width' or less.  (If 'break_long_words' is false,
    some lines may be longer than this.)  Chunks correspond roughly
    to words and the whitespace between them: each chunk is
    indivisible (modulo 'break_long_words'), but a line break can
    come between any two chunks.  Chunks should not have internal
    whitespace; ie. a chunk is either all whitespace or a "word".
    Whitespace chunks will be removed from the beginning and end of
    lines, but apart from that whitespace is preserved.
    """
    lines = []
    if self.width <= 0:
        raise ValueError("invalid width %r (must be > 0)" % self.width)
    if self.max_lines is not None:
        if self.max_lines > 1:
            indent = self.subsequent_indent
        else:
            indent = self.initial_indent
        if len(indent) + len(self.placeholder.lstrip()) > self.width:
            raise ValueError("placeholder too large for max width")

    # Arrange in reverse order so items can be efficiently popped
    # from a stack of chucks.
    chunks.reverse()

    while chunks:

        # Start the list of chunks that will make up the current line.
        # cur_len is just the length of all the chunks in cur_line.
        cur_line = []
        cur_len = 0

        # Figure out which static string will prefix this line.
        if lines:
            indent = self.subsequent_indent
        else:
            indent = self.initial_indent

        # Maximum width for this line.
        width = self.width - len(indent)

        # First chunk on line is whitespace -- drop it, unless this
        # is the very beginning of the text (ie. no lines started yet).
        if self.drop_whitespace and chunks[-1].strip() == '' and lines:
            del chunks[-1]

        while chunks:
            ## GPS - ignoring non-text
            last_chunk = chunks[-1]
            ignore_len = (
                last_chunk.startswith(cli_conf.STX)
                and last_chunk.endswith(cli_conf.ETX))
            l = 0 if ignore_len else len(last_chunk)

            # Can at least squeeze this chunk onto the current line.
            if cur_len + l <= width:
                cur_line.append(chunks.pop())
                cur_len += l

            # Nope, this line is full.
            else:
                break

        # The current line is full, and the next chunk is too big to
        # fit on *any* line (not just this one).
        if chunks and len(chunks[-1]) > width:
            self._handle_long_word(chunks, cur_line, cur_len, width)
            cur_len = sum(map(len, cur_line))

        # If the last chunk on this line is all whitespace, drop it.
        if self.drop_whitespace and cur_line and cur_line[-1].strip() == '':
            cur_len -= len(cur_line[-1])
            del cur_line[-1]

        if cur_line:
            if (self.max_lines is None or
                len(lines) + 1 < self.max_lines or
                (not chunks or
                 self.drop_whitespace and
                 len(chunks) == 1 and
                 not chunks[0].strip()) and cur_len <= width):
                # Convert current line back to a string and store it in
                # list of all lines (return value).
                lines.append(indent + ''.join(cur_line))
            else:
                while cur_line:
                    if (cur_line[-1].strip() and
                        cur_len + len(self.placeholder) <= width):
                        cur_line.append(self.placeholder)
                        lines.append(indent + ''.join(cur_line))
                        break
                    cur_len -= len(cur_line[-1])
                    del cur_line[-1]
                else:
                    if lines:
                        prev_line = lines[-1].rstrip()
                        if (len(prev_line) + len(self.placeholder) <=
                                self.width):
                            lines[-1] = prev_line + self.placeholder
                            break
                    lines.append(indent + self.placeholder.lstrip())
                break

    return lines


textwrap.TextWrapper._wrap_chunks = _wrap_text_chunks_only

ansi_escape = re.compile(r"\x1b[^m]*m")

def get_code_hl_tokens():
    code_hl_tokens = {}
    # replace code strs with tokens:
    for token_name, colour in cli_colour.TOKEN_NAME_TO_HL_COLOUR.items():
        if '.' not in token_name:  ## cope with Operator.Word as token_name
            code_hl_tokens[getattr(token, token_name)] = colour
        else:
            token_names = token_name.split('.')
            token2use = token
            for token_name in token_names:
                token2use = getattr(token2use, token_name)
            code_hl_tokens[token2use] = colour
    return code_hl_tokens

def style_ansi(raw_code):
    """
    Actual code_lines highlighting
    """
    lexer = get_lexer_by_name(cli_conf.PYTHON_LEXER_NAME)
    tokens = lex(raw_code, lexer)
    code_hl_tokens = get_code_hl_tokens()
    code_lines = []
    for my_token, text in tokens:
        if not text:
            continue
        colour = code_hl_tokens.get(my_token, cli_colour.CODE_COLOUR)
        code_lines.append(cli_colour.colourise(text, colour))
        logging.debug(my_token, colour)
    styled_ansi_code = ''.join(code_lines)
    return styled_ansi_code

def set_hr_widths(result):
    if cli_conf.HR_MARKER not in result:
        return result
    hrs = [line for line in result.splitlines() if cli_conf.HR_MARKER in line]
    for hr in hrs:
        hrf = hr.replace(
            cli_conf.HR_MARKER, cli_conf.HR_SEP * cli_conf.TERMINAL_WIDTH)
        result = result.replace(hr, hrf)
    return result

def apply_borders(text):
    border = cli_colour.colourise_low_vis(text[0].replace("-", "─"))
    text[0] = text[-1] = border

def split_blocks(text_block, width, n_cols, part_formatter=None):
    """
    Splits while multi-line blocks vertically (for large tables)
    """
    ts = []
    for line in text_block.splitlines():
        parts = []
        line = line.ljust(width, ' ')  ## make equal length
        # first part full width, others a bit indented:
        parts.append(line[:n_cols])
        scols = n_cols - 2
        # the txt_block_cut in low makes the whole secondary tables
        # low. which i find a feature:
        # if you don't want it remove the colourise(.., LOW_VIS_COLOUR)
        parts.extend(
            [
                (' '
                +
                cli_colour.colourise(
                    cli_conf.TEXT_BLOCK_CUT, cli_colour.LOW_VIS_COLOUR,
                    no_reset=True)
                + line[i : i + scols]
                )
                for i in range(n_cols, len(line), scols)
            ]
        )
        ts.append(parts)
    blocks = []
    for block_part_nr in range(len(ts[0])):
        tpart = []
        for lines_block in ts:
            tpart.append(lines_block[block_part_nr])
        if part_formatter:
            part_formatter(tpart)
        tpart[1] = cli_colour.colourise(tpart[1], cli_colour.H3_COLOUR)
        blocks.append("\n".join(tpart))
    text = 'n' + '\n'.join(blocks) + '\n'
    return text

def rewrap(el, text, indent, prefix, terminal_width):
    """
    Reasonably smart rewrapping checking punctuations.
    """
    cols = max(terminal_width - len(indent + prefix), 5)
    if el.tag == 'code' or len(text) <= cols:
        return text
    # this is a code replacement marker of markdown.py. Don't text split the
    # replacement marker:
    if text.startswith(cli_conf.STX) and text.endswith(cli_conf.ETX):
        return text
    new_lines = []
    for line in text.split('\n'):
        dedented = textwrap.dedent(line)
        new_lines.extend(textwrap.wrap(dedented, width=cols))  ## had to monkey patch textwrap to prevent code replacement markers messing up line wrapping
    new_text = '\n'.join(new_lines)
    return new_text

def clean_ansi(s):
    """
    If someone does not want the color foo
    """
    return ansi_escape.sub('', s)

def el2str(el):
    str2use = etree.tostring(el).decode('utf-8')
    return str2use

def get_text_if_inline_markup(el):
    """
    :return: html_str or None if not text with inline markup
    :rtype: str
    """
    el_str = el2str(el)
    ## strip tag
    html_str = (
        el_str.split(f'<{el.tag}', 1)[1].split('>', 1)[1].rsplit('>', 1)[0])
    ## start with another tagged child which is NOT in inlines?
    inlines = ['<a', '<em>', '<code>', '<strong>']
    has_inline_markup = False
    if html_str.startswith('<'):
        for inline in inlines:
            if html_str.startswith(inline):
                has_inline_markup = True
                break
    else:
        has_inline_markup = True
    if has_inline_markup:
        text_with_inline_markup = html_str
    else:
        text_with_inline_markup = None
    return text_with_inline_markup

def replace_links(el, html, link_display_type='it'):
    """
    Digging through inline "<a href=..."
    :return: links_list, and tag
    """
    parts = html.split('<a ')
    if len(parts) == 1:
        links_list = None
        return links_list, html
    links_list, cur_link = [], 0
    links = [l for l in el.getchildren() if "href" in l.keys()]
    if not len(parts) == len(links) + 1:
        ## contains an html element we don't support e.g. blockquote
        links_list = None
        return links_list, html
    cur = ''
    while parts:
        cur += parts.pop(0).rsplit("</a>")[-1]
        if not parts:
            break
        
        cur += cli_conf.LINK_START  ## indicating link formatting start
        # the 'a' xml element:
        link = links[cur_link]
        link_str = link.get('href', '')
        # bug in the markdown api? link el is not providing inlines!!
        # -> get them from the html:
        # cur += link.text or ''
        cur += parts[0].split('>', 1)[1].split('</a', 1)[0] or ''
        cur += cli_conf.LINK_END
        if link_display_type == "i":
            cur += cli_colour.colourise_low_vis(f"({link_str})")
        elif link_display_type != "h":  # inline table (it)
            # we build a link list, add the number like ① :
            cur += f"{chr(cli_conf.LINK_START_ORD + cur_link)} "
            links_list.append(link_str)
        else:
            pass  ## apparently we ignore h
        cur_link += 1
    return links_list, cur



from superhelp.helpers import filt_block_help
from .. import conf
from superhelp.helpers import get_aop_msg
from superhelp.gen_utils import get_nice_str_list, layout_comment as layout

DECORATOR_XPATH = (
    'descendant-or-self::decorator_list/Name '
    '| '
    'descendant-or-self::decorator_list/Call/func/Name'
    ' | '
    'descendant-or-self::decorator_list/Call/func/Attribute/value/Name'
)

@filt_block_help(xpath=DECORATOR_XPATH)
def decorator_overview(block_dets, *, repeat=False):
    """
    Look for decorators and explain some options for improving them.
    """
    decorator_els = block_dets.element.xpath(DECORATOR_XPATH)
    if not decorator_els:
        return None
    decorator_names = []
    for decorator_el in decorator_els:
        name = decorator_el.get('id')
        dec_list_el = decorator_el.xpath('ancestor::decorator_list')[0]
        attrib_els = dec_list_el.xpath('Call/func/Attribute')
        if len(attrib_els) == 1:
            namespace = attrib_els[0].get('attr')
            if namespace:
                name = f"{namespace}.{name}"
        decorator_names.append(name)

    dec_name_list = get_nice_str_list(decorator_names, quoter='`')
    plural = 's' if len(decorator_names) > 1 else ''
    summary = layout(f"""\
    ### Decorator{plural} used

    The code uses the decorator{plural}: {dec_name_list}.
    """)
    if not repeat:
        dec_dets = (
            layout("""\

            Decorators are a common and handy feature of Python. Using them is
            beginner-level Python and making them is intermediate-level Python.
            One tip - use functools.wraps to make the decorated function look
            like the original function e.g. have the same doc string.

            Decorators are well-documented elsewhere so there is no real advice
            to give here apart from recommending their use when appropriate.

            Here's an example of a simple decorator (`route`) making it easy to
            set up a web end point:
            """)
            +
            layout("""\
            from bottle import route

            @route('/hello')
            def hello():
                return "Hello World!"
            """, is_code=True)
            +
            layout("""\

            and here is an example of a simple, no argument decorator being
            created with `functool.wraps` applied:
            """)
            +
            layout('''\
            from functools import wraps

            def tweet(func):
                """
                Fake tweet original message after saying the message as
                before
                """
                @wraps(func)
                def wrapper(message):
                    func(message)
                    print(f"I'm tweeting the message {message}")
                return wrapper

            @tweet
            def say(message):
                """
                Print supplied message
                """
                print(message)

            say("sausage!")
            ''', is_code=True)
        )
        aop = get_aop_msg()
    else:
        dec_dets = ''
        aop = ''

    message = {
        conf.BRIEF: summary,
        conf.MAIN: summary + dec_dets,
        conf.EXTRA: aop,
    }
    return message



from superhelp.helpers import filt_block_help
from .. import code_execution, conf
from superhelp import gen_utils
from superhelp.gen_utils import get_nice_str_list, layout_comment as layout

def truncate_dict(input_dict):
    output_dict = {}
    for k, v in input_dict.items():
        if len(output_dict) >= conf.MAX_ITEMS_EVALUATED:
            break
        output_dict[k] = v
    return output_dict

ASSIGN_DICT_XPATH = 'descendant-or-self::Assign/value/Dict'

@filt_block_help(xpath=ASSIGN_DICT_XPATH)
def dict_overview(block_dets, *, repeat=False):
    """
    Look at assigned dictionaries e.g. location = {'country' 'New Zealand',
    'city': 'Auckland'}
    """
    dict_els = block_dets.element.xpath(ASSIGN_DICT_XPATH)
    plural = 'ies' if len(dict_els) > 1 else 'y'
    title = layout(f"""\
    ### Dictionar{plural} defined
    """)
    brief_desc_bits = []
    names_items, oversized_msg = code_execution.get_collections_dets(
        dict_els, block_dets,
        collection_plural='dictionaries', truncated_items_func=truncate_dict)
    for name, items in names_items:
        if items is None:
            brief_desc_bits.append(layout(f"""\
            `{name}` is a dictionary. Unable to evaluate items.
            """))
        else:
            brief_desc_bits.append(layout(f"""\

            `{name}` is a dictionary with {gen_utils.int2nice(len(items))} items
            (i.e. {gen_utils.int2nice(len(items))} mappings).
            """))
    brief_desc = ''.join(brief_desc_bits)
    if not repeat:
        dict_def = layout("""\
        Dictionaries map keys to values.
        """)
        workhorses = layout("""\

        Dictionaries, along with lists, are the workhorses of Python data
        structures.
        """)
        keys_and_vals = layout("""\
        Keys are unique but values can be repeated.
        """)
        dict_desc_bits = []
        first_name = None
        for name, items in names_items:
            if not items:
                dict_desc_bits.append(layout(f"""\
                `{name}` is a dictionary. Unable to evaluate items.
                """))
            else:
                if not first_name:
                    first_name = name
                empty_dict = (len(items) == 0)
                if empty_dict:
                    dict_desc_bits.append(layout(f"""\
                    `{name}` is an empty dictionary.
                    """))
                else:
                    plural = '' if len(items) == 1 else 's'
                    dict_desc_bits.append(layout(f"""\

                    `{name}` is a dictionary with {gen_utils.int2nice(len(items))}
                    item{plural} (i.e. {gen_utils.int2nice(len(items))}
                    mapping{plural}). In this case, the keys are:
                    {list(items.keys())}. We can get the keys using the
                    `.keys()` method e.g. `{name}`.`keys()`. The values are
                    {list(items.values())}. We can get the values using the
                    `.values()` method e.g. `{name}`.`values()`.
                    """))
        if first_name is None:
            first_name = 'demo_dict'
        main_dict_desc = ''.join(dict_desc_bits)
        general = (
            layout(f"""

            It is common to iterate through the key-value pairs of a dictionary.
            This can be achieved using the dictionary's `.items()` method. E.g.
            """)
            +
            layout(f"""\

            ## k, v is conventional, and OK in a hurry, but readable names
            ## are probably better for code you're going to maintain
            for k, v in {first_name}.items():
                print(f"key {{k}} maps to value {{v}}")
            """, is_code=True)
            +
            layout("""
            Keys are unique but values can be repeated. For example:
            """)
            +
            layout(f"""
            country2car = {{'Japan': 'Toyota', 'Sweden': 'Volvo'}}  ## OK - all keys are unique
            country2car = {{'Japan': 'Toyota', 'Japan': 'Honda'}}  ## Oops - the 'Japan' key is repeated
            """, is_code=True)
            +
            layout(f"""

            In which case a better structure might be to have each 'value' being
            a list e.g.
            """)
            +
            layout(f"""
            country2cars = {{'Japan': ['Toyota', 'Honda'], 'Sweden': ['Volvo']}}  ## OK - all keys are unique
            """, is_code=True)
        )
        mighty_dict = layout("""\

        Python dictionaries (now) keep the order in which items are added.

        They are also super-efficient and fast. The two presentations to watch
        are by living treasure Brandon Rhodes:

        1. The Dictionary Even Mightier -
        <https://www.youtube.com/watch?v=66P5FMkWoVU>
        2. The Mighty Dictionary - <https://www.youtube.com/watch?v=oMyy4Sm0uBs>
        """)
    else:
        dict_def = ''
        workhorses = ''
        keys_and_vals = ''
        main_dict_desc = brief_desc
        general = ''
        mighty_dict = ''

    message = {
        conf.BRIEF: (title + oversized_msg + dict_def + brief_desc
            + keys_and_vals + workhorses),
        conf.MAIN: (
            title + oversized_msg + main_dict_desc + workhorses + general),
        conf.EXTRA: mighty_dict,
    }
    return message

def get_key_type_names(items):
    key_type_names = sorted(set(
        [type(item).__name__ for item in items]
    ))
    key_type_nice_names = [
        conf.TYPE2NAME.get(key_type, key_type)
        for key_type in key_type_names]
    return key_type_names, key_type_nice_names

@filt_block_help(xpath=ASSIGN_DICT_XPATH, warning=True)
def mixed_key_types(block_dets, *, repeat=False):
    """
    Warns about dictionaries with mix of string and integer keys.
    """
    dict_els = block_dets.element.xpath(ASSIGN_DICT_XPATH)
    mixed_names = []
    names_items, oversized_msg = code_execution.get_collections_dets(
        dict_els, block_dets,
        collection_plural='dictionaries', truncated_items_func=truncate_dict)
    for name, items in names_items:
        if not items:
            continue
        key_type_names, _key_type_nice_names = get_key_type_names(items)
        bad_key_type_combo = (
            conf.INT_TYPE in key_type_names and conf.STR_TYPE in key_type_names)
        if not bad_key_type_combo:
            continue
        mixed_names.append(name)
    if not mixed_names:
        return None

    title = layout("""\
    ### Mix of integer and string keys in dictionary
    """)
    multiple = len(mixed_names) > 1
    if multiple:
        nice_str_list = get_nice_str_list(mixed_names, quoter='`')
        mixed_warning = layout(f"""

        {nice_str_list} have keys include a mix of strings and integers - which
        is probably a bad idea.
        """)
    else:
        mixed_warning = layout(f"""

        `{name}`'s keys include both strings and integers which is probably a
        bad idea.
        """)
    if not repeat:
        one_vs_1 = layout("""\

        For example, if you have both 1 and "1" as keys in a dictionary (which
        is allowed because they are not the same key) it is very easy to get
        confused and create Hard To Find™ bugs. You _might_ not regret it but
        you probably will ;-).
        """)
    else:
        one_vs_1 = ''

    message = {
        conf.BRIEF: title + oversized_msg + mixed_warning,
        conf.MAIN: title + oversized_msg + mixed_warning + one_vs_1,
    }
    return message



from collections import defaultdict, namedtuple
from pathlib import Path
import re
from subprocess import run, PIPE
import sys

from ..helpers import snippet_str_help
from .. import conf, lint_conf
from ..gen_utils import (get_nice_str_list, get_os_platform,
    layout_comment as layout, make_open_tmp_file)

prog = re.compile(lint_conf.LINT_PATTERN, flags=re.VERBOSE)  # @UndefinedVariable

MsgDets = namedtuple('MsgDets', 'msg, line_no')

MISC_ISSUES_TITLE = layout("""\
    #### Misc lint issues
    """)

def _store_snippet(snippet):
    """
    At least one test (E501 line too long) only triggered if a trailing newline.
    Note - if more than one newline we trigger W391 (blank line at end of file)
    so an rstrip('\n') needed.
    Having done this need to deactivate W292 (blank line at end of file) LOL
    """
    with make_open_tmp_file(conf.SNIPPET_FNAME, mode='w') as tmp_dets:
        _superhelp_tmpdir, tmp_fh, fpath = tmp_dets
        tmp_fh.write(snippet.rstrip('\n') + '\n')
    return fpath

def _get_env_flake8_fpath():
    """
    Perhaps we are in a virtual environment which hides flake8 from which /
    where commands. Worth a try.
    """
    bin_path = Path(sys.executable).parent
    flake8_fpath = str(bin_path / 'flake8')
    return flake8_fpath

def _get_flake8_fpath():
    os_platform = get_os_platform()
    if os_platform == conf.WINDOWS:
        """
        The Joys of Where

        The "where" statement won't always point to an actual executable -
        sometimes it finds a stub which, when executed, opens up an MS app store
        LOL FAIL. Sometimes refers to two places - one on a C: drive and another
        on a D: drive!!
        """
        which_statement = 'where'
    else:
        which_statement = 'which'
    args = [which_statement, 'flake8']
    res = run(args=args, stdout=PIPE)
    flake8_fpath = str(res.stdout, encoding='utf-8').strip()
    if not flake8_fpath:
        flake8_fpath = _get_env_flake8_fpath()
    return flake8_fpath

def _get_flake8_results(fpath):
    flake8_fpath = _get_flake8_fpath()
    args = [flake8_fpath, str(fpath)]
    if lint_conf.IGNORED_LINT_RULES:
        ignored = ','.join(lint_conf.IGNORED_LINT_RULES)
        args.append(f"--ignore={ignored}")
    res = run(args=args, stdout=PIPE)
    return res

def _msg_type_to_placeholder_key(msg_type):
    return f"{msg_type}_placeholder"

def _msg_type_to_placeholder(msg_type):
    placeholder_key = _msg_type_to_placeholder_key(msg_type)
    placeholder = '{' + placeholder_key + '}'  ## ready for .format()
    return placeholder

def _get_msg_type_and_dets(lint_regex_dicts):
    """
    Gather MsgDets named tuples by message type e.g. all the E123s together.
    Some lint message types are consolidated as per
    lint_conf.CONSOLIDATE_MSG_TYPE.

    Each message type may have messages for multiple lines. These lines may have
    the same or different messages. E.g. "line too long (91 > 79 characters)" vs
    "line too long (100 > 79 characters)".

    We may also want to supplement / replace the standard messages with special
    messages. When replacing a message, we automatically want to consolidate to
    one message. If not replacing the message, we want our special content to
    appear after the unconsolidated messages for the (possibly consolidated)
    message type. We use placeholders to make this possible. These are replaced
    later on when making message level-specific messages (brief and main only).

    Message consolidation: for example, "line too long (91 > 79 characters)" and
    "line too long (100 > 79 characters)" become "line too long", and there will
    be one message with multiple line numbers listed rather than multiple
    messages, one per line.

    Note: just because a message type is consolidated doesn't mean its messages
    will be consolidated or vice versa. They are independent transformations.

    :param list lint_regex_dicts: list of dicts - one per regex.
     Basically a list of pulled apart lint messages.
    :return: dict of message types as keys (possibly consolidated e.g.
     E123-9 -> line continuation message type) and MsgDets tuples as values.
    :rtype: dict
    """
    msg_type_and_dets = defaultdict(list)
    for lint_regex_dict in lint_regex_dicts:
        raw_msg_type = lint_regex_dict[conf.LINT_MSG_TYPE]
        msg_type = lint_conf.consolidated_msg_type(raw_msg_type)
        msg = layout(lint_regex_dict[conf.LINT_MSG])
        line_no = int(lint_regex_dict[conf.LINT_LINE_NO])
        msg_dets = MsgDets(msg, line_no)
        msg_type_and_dets[msg_type].append(msg_dets)
    return msg_type_and_dets

def _get_msg_line(msgs_dets):
    """
    Get a single message string. Handle details for message type across all
    lines and specific messages.

    :rtype: string
    :return: a single line
    """
    msg_type_details = []
    msg_line_nos = defaultdict(set)
    for msg_dets in msgs_dets:
        msg_line_nos[msg_dets.msg].add(msg_dets.line_no)
    for msg, line_nos in msg_line_nos.items():
        plural = 's' if len(line_nos) > 1 else ''
        nice_lines = get_nice_str_list(sorted(line_nos))
        msg2use = msg[0].upper() + msg[1:]  ## note - .capitalize() lower cases the remaining letters
        msg2use = msg2use.replace('>', '\>').replace('<', '\<')
        msg_type_details.append(f"{msg2use} (line{plural}:{nice_lines})")
    msg_line = layout('; '.join(msg_type_details))
    return msg_line

def _get_unfinished_messages(msg_type_and_dets):
    """
    Get list of unfinished messages. May contain placeholders that need to be
    replaced.
    """
    unfinished_msgs = []
    already_supplemented = set()
    for msg_type, msgs_dets in msg_type_and_dets.items():
        unfinished_msg = _get_msg_line(msgs_dets)
        generic_msg_type = msg_type not in lint_conf.CUSTOM_LINT_MSGS
        if generic_msg_type:
            unfinished_msg = '* ' + unfinished_msg.lstrip('\n')  ## bullet points for generic messages (custom messages have full layout treatment e.g. code highlighting)
        ## add supplementary line?
        supplement_configured = msg_type in lint_conf.CUSTOM_LINT_MSGS
        if msg_type not in already_supplemented and supplement_configured:
            msg_placeholder = _msg_type_to_placeholder(msg_type)
            unfinished_msg = (
                msg_placeholder
                    + '\n\nDetails: '
                    + unfinished_msg.lstrip('\n')
                )  ## will be replaced by appropriate level message in brief and main
            already_supplemented.add(msg_type)
        unfinished_msgs.append(unfinished_msg)
    return unfinished_msgs

def _get_extra_msg(msg_type_and_dets):
    msg_types_used = msg_type_and_dets.keys()
    extra_lines = []
    for msg_type in msg_types_used:
        msgs = lint_conf.CUSTOM_LINT_MSGS.get(msg_type)
        if not msgs:
            continue
        extra_lines.append(msgs.extra)
    extra_msg = '\n\n'.join(extra_lines)
    return extra_msg

def _heading_sort_order(msg_line):
    """
    Custom first (i.e. generic False first), then by message i.e. by title
    """
    custom = ' #### ' in msg_line[: 10]  ## Custom all start with #### ...
    return (not custom, msg_line)  ## for sorting - False is first

def _get_final_msgs_for_level(msgs_for_msg_level):
    """
    Can only add Misc heading once everything is sorted by actual message
    content (i.e. placeholders replaced and then sorted)
    """
    final_msgs_for_level = []
    has_generic_header = False
    for msg_for_msg_level in msgs_for_msg_level:
        starting_misc = '* ' in msg_for_msg_level[:5]
        if not has_generic_header and starting_misc:
            msgs2extend = [MISC_ISSUES_TITLE, msg_for_msg_level]
            has_generic_header = True
        else:
            msgs2extend = [msg_for_msg_level, ]
        final_msgs_for_level.extend(msgs2extend)
    return final_msgs_for_level

def get_lint_messages_by_level(raw_lint_feedback_str):
    """
    Gets lists of lint messages grouped by message level (brief, main, extra).

    Each message in a list is for a particular error / warning type e.g. E501.

    A message might be consolidated (e.g. foo has prob; bar has prob etc) or be
    a composite message (e.g. various have prob). It might be followed by
    supplementary content.

    :param str raw_lint_feedback_str: feedback as received from the linter as
     stdout (told you it was raw ;-))
    :return: brief_msg, main_msg, extra_msg
    :rtype list
    """
    ## split into lines
    raw_lint_lines = (
        str(raw_lint_feedback_str, encoding='utf-8').strip().split('\n'))
    lint_lines = [line.strip() for line in raw_lint_lines if line.strip()]
    ## REGEX the lines
    lint_regex_dicts = []
    for lint_line in lint_lines:
        result = prog.match(lint_line)
        lint_regex_dicts.append(result.groupdict())
    ## misc
    msg_type_and_dets = _get_msg_type_and_dets(lint_regex_dicts)
    unfinished_messages = _get_unfinished_messages(msg_type_and_dets)
    ## replace placeholders with level-appropriate messages
    ## we can finally sort messages within a brief / main message level!
    lint_msgs = []
    for msg_level in ['brief', 'main']:
        msg_type_to_msg_for_level = {
            _msg_type_to_placeholder_key(msg_type): getattr(msg_dets, msg_level)
            for msg_type, msg_dets in lint_conf.CUSTOM_LINT_MSGS.items()
        }
        msgs_for_level = [
            unfinished_message.format(**msg_type_to_msg_for_level)  ## maps msg_type e.g. E501 to e.g. "Line too long", E666_msg='Code is generally evil'
            for unfinished_message in unfinished_messages]
        msgs_for_level.sort(key=_heading_sort_order)
        ## Can only add Misc heading once everything is sorted by actual message
        ## content (i.e. placeholders replaced and then sorted)
        final_msgs_for_level = _get_final_msgs_for_level(msgs_for_level)
        final_msg_for_level = '\n\n'.join(final_msgs_for_level)
        lint_msgs.append(final_msg_for_level)
    ## handle extra - see which placeholders are present and provide extra for those only
    extra_msg = _get_extra_msg(msg_type_and_dets)
    lint_msgs.append(extra_msg)
    return lint_msgs

@snippet_str_help(warning=True)
def lint_snippet(snippet):
    """
    Look for "lint" as defined by flake8 linter and share the results.
    """
    if not conf.INCLUDE_LINTING:  ## disabled when testing for speed reasons
        return None
    fpath = _store_snippet(snippet)
    res = _get_flake8_results(fpath)
    if not res.stdout:
        return None

    title = layout("""\
    ### Python code issues (found by flake8 linter)
    """)
    linting = layout("""\

    "Linters" are software tools. They detect everything from trivial style
    mistakes of no consequence to program behaviour through to show-stopper
    syntax errors.

    Software developers can be notoriously fussy about the smallest details of
    code styling and a linter can not only detect actual errors in code it can
    also prevent developers becoming completely distracted by trivial irritants.
    Distracted developers miss real issues with programs so it can be of
    practical importance to pick up the "small stuff". Plus it enables teams of
    programmers to work on the same code base without spending all their time
    restyling each other's code and arguing about
    "standards".
    """)
    findings = layout("""\
    Here is what the linter reported about your snippet.
    """)
    brief_msg, main_msg, extra_msg = get_lint_messages_by_level(
        raw_lint_feedback_str=res.stdout)
    obviousness = layout("""\
    #### Good code is simple enough to reason about

    Linting is especially useful for an interpreted language like Python because
    there is no compiler to pick up "lint" errors. Linting is no substitute for
    unit testing though. And neither are a substitute for writing readable code
    that can be reasoned about with confidence - the single best protection
    against code not doing what it is meant to do. The goal should be code where
    there is obviously nothing wrong rather than code where there nothing
    obviously wrong.

    > "There are two ways of constructing a software design. One way is to make
    it so simple that there are obviously no deficiencies. And the other way is
    to make it so complicated that there are no obvious deficiencies." C.A.R.
    Hoare
    """)

    message = {
        conf.BRIEF: title + findings + brief_msg,
        conf.MAIN: title + linting + findings + main_msg,
        conf.EXTRA: obviousness + extra_msg,
    }
    return message



from collections import defaultdict

from superhelp.helpers import filt_block_help
from ..ast_funcs import assigned_num_els_from_block
from .. import code_execution, conf, name_utils
from superhelp.gen_utils import get_nice_str_list, layout_comment as layout

ASSIGN_VAL_XPATH = 'descendant-or-self::Assign/value'

@filt_block_help(xpath=ASSIGN_VAL_XPATH)
def num_overview(block_dets, *, repeat=False):
    """
    Get general advice about assigned numbers e.g. var = 123
    """
    num_els = assigned_num_els_from_block(block_dets.element)
    val_types = defaultdict(list)
    has_num = False
    type_firsts = {}
    for num_el in num_els:
        name_dets = name_utils.get_assigned_name(num_el)
        try:
            val = code_execution.get_val(
                block_dets.pre_block_code_str, block_dets.block_code_str,
                name_dets.name_type, name_dets.name_details, name_dets.name_str)
        except KeyError:
            continue
        else:
            val_type = type(val).__name__
            val_types[val_type].append(name_dets.name_str)
            if not type_firsts.get(val_type):
                type_firsts[val_type] = val
            has_num = True
    if not has_num:
        return None

    title = layout("""\
    ### Number details
    """)
    names_msg_bits = []
    for val_type, names in val_types.items():
        names_text = get_nice_str_list(names, quoter='`')
        if len(names) == 1:
            names_msg_bits.append(layout(f"""\

            {names_text} is a number - specific type `{val_type}`.
            """))
        else:
            names_msg_bits.append(layout(f"""\

            {names_text} are numbers - specific type `{val_type}`.
            """))
    names_msg = ''.join(names_msg_bits)
    if not repeat:
        specifics_bits = []
        for val_type, names in val_types.items():
            first_name = names[0]
            val = type_firsts[val_type]
            specific_comment = None
            if val_type == conf.INT_TYPE:
                specific_comment = layout(f"""\

                Integers are counting numbers and include 0 and negative numbers
                e.g. -2

                If you need a float instead of an integer use the float
                function.

                e.g. float({first_name}) which returns {float(val)}
                """)
            elif val_type == conf.FLOAT_TYPE:
                specific_comment = layout(f"""\

                Floats are used when decimal places are required.

                If you need an integer instead of a float use the int function.

                e.g. int({first_name}) which returns {int(val)}
                """)
            if specific_comment:
                specifics_bits.append(specific_comment)
        specifics = ''.join(specifics_bits)
        if conf.FLOAT_TYPE in val_types:
            floats = layout(f"""\

            Floats, or floating point numbers, are stored in computers as binary
            fractions. "Unfortunately, most decimal fractions cannot be
            represented exactly as binary fractions. A consequence is that, in
            general, the decimal floating-point numbers you enter are only
            approximated by the binary floating-point numbers actually stored in
            the machine." For more information, read the rest of
            <https://docs.python.org/3/tutorial/floatingpoint.html>. It is
            really interesting - honest!
            """)
        else:
            floats = ''
    else:
        specifics = ''
        floats = ''

    message = {
        conf.BRIEF: title + names_msg + specifics,
        conf.EXTRA: floats,
    }
    return message



import argparse
import logging
import os

"""
Do not use relative imports (dot notation) in this module. Relative importing
would be an option if we only wanted to run this module as part of superhelp as
a library but not if we also want to run it as a script. Running this module as
a script is very useful for development purposes from within the IDE - it allows
full step-by-step debugging etc. Absolute imports make this possible because
statements like:

    from superhelp import conf

are successful both when helper.py is run as part of a library module and when
it is run as a script. How? When run as part of a library the import statement
functions as an absolute import relating everything to the root folder of the
library. As a script, the import statement runs superhelp as an installed site
package. Obviously, for this library import to work we have to have superhelp
installed as a site package.

To reference the existing superhelp code in the IDE when importing superhelp as
a library the superhelp package must have been installed in --edit mode which
places a link in the standard site-packages folder pointing to this set of
files.

Because superhelp references this module in its __init__ we pass through this
module twice when we run this module as a script. The first time starts when we
run the script. As we proceed through the script line-by-line, we hit the import
statement and then import superhelp as a library. That imports the entire module
again (the library links to the exact same file so it can reference the `this`
function) and we then continue through the rest of the module.

In tests we should import superhelp as a library, i.e. no relative importing, so
tests can be run from anywhere.
"""
from superhelp import conf, gen_utils, helpers, messages
from superhelp.displayers import cli_displayer, html_displayer, md_displayer

logging.basicConfig(
    level=conf.LOG_LEVEL,
    format='%(asctime)s %(levelname)-8s %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S')

helpers.load_helpers()

def _get_displayer_module(output):
    ARG2DISPLAYER = {
        'html': html_displayer,
        'cli': cli_displayer,
        'md': md_displayer,
    }
    displayer_module = ARG2DISPLAYER.get(output)
    if displayer_module is None:
        logging.info("Display is currently suppressed - please supply "
            "a displayer if you want advice displayed")
    return displayer_module

def _run_display(code, displayer_module, messages_dets, *,
        file_path=None, output='html', detail_level=conf.EXTRA,
        warnings_only=False, in_notebook=False,
        multi_script=False, multi_block=False,
        theme_name=None):
    if len(str(file_path)) > conf.MAX_FILE_PATH_IN_HEADING:
        n_chars2_keep = conf.MAX_FILE_PATH_IN_HEADING - 4
        file_path = '...' + str(file_path)[-n_chars2_keep:]
    kwargs = {'multi_script': multi_script}
    if output == 'cli':
        kwargs.update({'theme_name': theme_name})
    elif output == 'html':
        kwargs.update(
            {'in_notebook': in_notebook, 'multi_script': multi_script})
    else:
        pass
    deferred_display = displayer_module.display(code, file_path,
        messages_dets, detail_level=detail_level,
        warnings_only=warnings_only, multi_block=multi_block,
        **kwargs)
    return deferred_display

def get_code_help(code, *, file_path=None,
        output='html', detail_level=conf.EXTRA,
        warnings_only=False, in_notebook=False,
        multi_script=False, theme_name=None, repeat_set=None):
    if multi_script and in_notebook:
        raise Exception("Notebooks should only ever run on snippets not on "
            "multiple scripts")
    if code.strip() == 'import community':
        messages_dets = messages.get_community_message(code)
        multi_block = False
    elif all([word in code for word in conf.XKCD_WARNING_WORDS]):
        messages_dets = messages.get_xkcd_warning(code)
        multi_block = False
    else:
        try:
            messages_dets, multi_block = messages.get_snippet_dets(
                code, warnings_only=warnings_only, repeat_set=repeat_set)
        except Exception as e:
            messages_dets = messages.get_error_messages_dets(e, code)
            multi_block = False
    displayer_module = _get_displayer_module(output)
    if displayer_module:
        deferred_display = _run_display(code, displayer_module, messages_dets,
            file_path=file_path, output=output, detail_level=detail_level,
            warnings_only=warnings_only, in_notebook=in_notebook,
            multi_script=multi_script, multi_block=multi_block,
            theme_name=theme_name)
    else:
        deferred_display = None
    return deferred_display

def get_script_help(file_path, *, output='html', detail_level=conf.EXTRA,
        warnings_only=False, in_notebook=False,
        multi_script=False, theme_name=None, repeat_set=None):
    with open(file_path) as f:
        code = f.read()
    code = code.strip('\n')
    ## prevent infinite recursion where superhelp executes script calling superhelp which in turn would etc etc
    code = (code  ## only fixing simple cases - if people try harder they _will_ be able to break everything ;-)
        .replace('import superhelp', '# import superhelp')
        .replace('\nsuperhelp.this(', '\n# superhelp.this(')
        .replace('from superhelp import this', '# from superhelp import this')
        .replace('\nthis(', '\n# this(')
    )
    get_code_help(code, file_path=file_path,
        output=output, detail_level=detail_level,
        warnings_only=warnings_only, in_notebook=in_notebook,
        multi_script=multi_script, theme_name=theme_name, repeat_set=repeat_set)

def _get_file_paths(project_path, exclude_folders):
    """
    Very easy to end up with far too many modules to process e.g. if
    inadvertently looking at every module inside the site packages in a virtual
    env ;-).
    """
    file_paths = []
    for root, dirs, files in os.walk(project_path, topdown=True):
        dirs[:] = [d for d in dirs if d not in exclude_folders]
        py_files = [os.path.join(root, file)
            for file in files if file.endswith('.py')]
        file_paths.extend(py_files)
    if len(file_paths) > conf.MAX_PROJECT_MODULES:
        raise Exception(
            f"Too many modules to process - {len(file_paths):,}")
    return file_paths

def get_help(code=None, *,
        file_path=None, project_path=None, exclude_folders=None,
        output='html', detail_level=conf.EXTRA,
        warnings_only=False, in_notebook=False, theme_name=None):
    """
    If a snippet of code supplied, get help for that. If not, try file_path and
    use that instead. And if not that, try project_path. Finally use the default
    snippet.

    Only getting deferred display from displayers when looking at code snippets.

    :param str code: (optional) snippet of valid Python code to get help for.
     If None will try the file_path, then the project_path, and finally the
     default snippet.
    :param str file_path: (optional) file path containing Python code
    :param str project_path: (optional) path to project containing Python code
    :param list exclude_folders: may be crucial if setting project_path e.g. to
     avoid processing all python scripts in a virtual environment folder
    :param str output: type of output e.g. 'html', 'cli', 'md'. Defaults to
     'html'.
    :param str detail_level: e.g. 'Brief', 'Main', 'Extra'
    :param bool warnings_only: if True only displays warnings
    :param bool in_notebook: if True might change way display happens e.g. HTML
     not sent to browser but returned for display by notebook itself
    :param str theme_name: currently only needed by the CLIC displayer (to
     handle dark and light terminal themes)
    """
    repeat_set = set()  ## mutates as we hand it around to keep track of repeats
    if code:
        code = code.strip('\n')
        deferred_display = get_code_help(code, file_path=None,
            output=output, detail_level=detail_level,
            warnings_only=warnings_only, in_notebook=in_notebook,
            multi_script=False, theme_name=theme_name, repeat_set=repeat_set)
        if not deferred_display:
            return
        if (in_notebook and output == 'html'):
            return deferred_display  ## assumed notebook always HTML and we need to pass on the HTML string to the jupyter notebook
        else:
            raise Exception(
                "Only a notebook with html output should use deferred display")
    elif file_path:
        get_script_help(file_path,
            output=output, detail_level=detail_level,
            warnings_only=warnings_only, in_notebook=in_notebook,
            multi_script=False, theme_name=theme_name, repeat_set=repeat_set)
    elif project_path:
        file_paths = _get_file_paths(project_path, exclude_folders)
        for file_path in file_paths:
            get_script_help(file_path, output=output, detail_level=detail_level,
                warnings_only=warnings_only, in_notebook=in_notebook,
                multi_script=True, theme_name=theme_name, repeat_set=repeat_set)
        if output == 'html':
            gen_utils.open_output_folder()
    else:
        code = conf.TEST_SNIPPET
        logging.info("Using default snippet because no code provided")
        get_code_help(code, file_path=None,
            output=output, detail_level=detail_level,
            warnings_only=warnings_only, in_notebook=in_notebook,
            multi_script=False, theme_name=theme_name, repeat_set=repeat_set)

def this(*, output='html', detail_level=conf.EXTRA,
        warnings_only=False, file_path=None, theme_name='dark'):
    """
    Get SuperHELP output on the file_path Python script.

    Yes - this messes up the ability to "import this" later because I've shaded
    the name "this" in this namespace by naming this function, err ..., "this".
    But it was considered of paramount importance to give users a simple and
    easy-to-remember superhelp.this() interface.

    :param str output: type of output e.g. 'html', 'cli', 'md'. Defaults to
     'html'.
    :param str detail_level: e.g. 'Brief', 'Main', 'Extra'
    :param bool warnings_only: if True only displays warnings
    :param str / Path file_path: full path to script location. Only needed if
     SuperHELP is unable to locate the script by itself. Usually should be
     __file__ (note the double underscores on either side of file).
    :param str theme_name: currently only needed by the CLIC displayer (to
     handle dark and light terminal themes)
    """
    if not file_path:
        file_path = gen_utils.get_introspected_file_path()
    get_help(code=None, file_path=file_path, project_path=None,
        output=output, detail_level=detail_level, warnings_only=warnings_only,
        in_notebook=False, theme_name=theme_name)

def shelp():
    """
    To get help

    $ shelp -h
    """
    default_output = conf.OUTPUT
    ## don't use type=list ever https://stackoverflow.com/questions/15753701/argparse-option-for-passing-a-list-as-option
    parser = argparse.ArgumentParser(
        description='Superhelp - Help for Humans!')
    parser.add_argument('-c', '--code', type=str,
        required=False,
        help=("Python code - usually only a line or snippet. "
            "An alternative to --file-path and --project-path"))
    parser.add_argument('-f', '--file-path', type=str,
        required=False,
        help=("File location of a line, snippet, or script of Python code. "
            "An alternative to --code and --project-path"))
    parser.add_argument('-p', '--project-path', type=str,
        required=False,
        help=("Project folder containing all the modules you want help on. "
            "An alternative to --code and --file-path"))
    parser.add_argument('-e', '--exclude-folders', type=str,
        nargs='*', help=("If using -p / --project-path you probably need to "
            "supply to exclude modules in storage folders or a virtual env "
            "folder e.g. --exclude store env back_ups misc"))
    parser.add_argument('-d', '--detail-level', type=str,
        required=False, default='Extra',
        help="What level of detail do you want? Brief, Main, or Extra?")
    parser.add_argument('-o', '--output', type=str,
        required=False, default=default_output,
        help="How do you want your help shown? html, cli, md, etc")
    parser.add_argument('-w', '--warnings-only', action='store_true',
        default=False,
        help="Show warnings only")
    parser.add_argument('-t', '--theme', type=str,
        required=False, default='dark',
        help=("Select an output theme - currently only affects cli output "
            "option. 'dark' or 'light'"))
    parser.add_argument('-a', '--advice-list', action='store_true',
        default=False,
        help="List available advice")
    args = parser.parse_args()
    if args.advice_list:
        print("\n======================================")
        print("Specific help available from SuperHELP")
        print("======================================\n")
        helper_comments = helpers.get_helper_comments()
        num_width = len(str(len(helper_comments)))
        for n, (comment, source, warning) in enumerate(helper_comments, 1):
            print(f"{n:>{num_width}}) {warning}{comment} ({source})")
        return
    sources = [
        src for src in (args.code, args.file_path, args.project_path) if src]
    if len(sources) > 1:
        print(
            "The code you want help on can only be identified in one way. "
            "Either supply code by:"
            "\n1) supplying it directly using -c / --code "
            "(usually for smaller snippets of Python)"
            "\n2) or by referring to a Python script using -f / --file-path "
            "\n3) or by referring to a folder of Python modules using "
            "-p / --project-path"
        )
        return
    logging.debug(args)
    output = args.output if conf.SHOW_OUTPUT else None
    theme_name = args.theme
    get_help(args.code,
        file_path=args.file_path,
        project_path=args.project_path, exclude_folders=args.exclude_folders,
        output=output, theme_name=theme_name,
        detail_level=args.detail_level, warnings_only=args.warnings_only,
        in_notebook=False
    )

if __name__ == '__main__':
#     get_help(file_path='/home/g/projects/superhelp/superhelp/store/misc_scripts/get_digicam_videos_off.py') #, output=None)
    shelp()



from setuptools import setup, find_packages  # @UnresolvedImport
from codecs import open
from os import path

__version__ = '1.0.4'

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# get the dependencies and installs
with open(path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    all_reqs = f.read().split('\n')

install_requires = [x.strip() for x in all_reqs if 'git+' not in x]
dependency_links = [x.strip().replace('git+', '') for x in all_reqs if x.startswith('git+')]

setup(
    name='superhelp',
    version=__version__,
    description='SuperHELP - Help for Humans!',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/grantps/superhelp',
    license='MIT',
    classifiers=[
      'Development Status :: 4 - Beta',
      'Intended Audience :: Developers',
      'Programming Language :: Python :: 3',
      'Programming Language :: Python :: 3.6',
      'Programming Language :: Python :: 3.7',
      'Programming Language :: Python :: 3.8',
      'Programming Language :: Python :: 3.9',
      'Operating System :: OS Independent',
    ],
    keywords='',
    packages=find_packages(exclude=['docs', 'tests*']),
    include_package_data=True,
    author='Grant Paton-Simpson',
    install_requires=install_requires,
    dependency_links=dependency_links,
    author_email='grant@p-s.co.nz',
    python_requires='>=3.6',
    entry_points = {
        'console_scripts': [
            'shelp=superhelp:shelp',  ## using argparse to allow arguments
        ]
    },
)



BOLD = '\u001b[1m'
UNDERLINE = '\u001b[4m'
REVERSED = '\u001b[7m'

PYTHON_LEXER_NAME = 'Python3'

HEADER_TAGS = [f"h{n}" for n in range(1, 9)]  ## h1, h2, ...

TERMINAL_WIDTH = 80  ## just hard-wiring this

ADMON_START = '!!! '

BQUOTE_PREFIX = '|'
LIST_PREFIX = '- '
CODE_PREFIX = '| '
TEXT_BLOCK_CUT = '✂'
HR_SEP = '─'

## nesting indentation
N_LEFT_INDENT = 2
LEFT_INDENT = ' ' * N_LEFT_INDENT

# markers: tab is 09, omit that
CODE_START, CODE_END = '\x07', '\x08'
STRONG_START, STRONG_END = '\x16', '\x10'
EMPH_START, EMPH_END = '\x11', '\x12'
LINK_START, LINK_END = '\x17', '\x18'
HR_MARKER = '\x15'
HR_ENDS = '◈'

CODE_BOUNDS = (CODE_START, CODE_END)
STRONG_BOUNDS = (STRONG_START, STRONG_END)
EMPH_BOUNDS = (EMPH_START, EMPH_END)
LINK_BOUNDS = LINK_START, LINK_END

TAG2BOUNDS = {
    '<code>': CODE_BOUNDS,
    '<strong>': STRONG_BOUNDS,
    '<em>': EMPH_BOUNDS,
}

## https://en.wikipedia.org/wiki/Control_character
STX = '\x02'  ## Start of transmission of non-data characters
ETX = '\x03'  ## End of transmission

BADLY_PARSED_UNDERSCORE = f'{STX}95{ETX}'

LINK_START_ORD = ord("①")
